"""Tooled Python worker to create a transcription on Arkindex elements"""

import logging

from arkindex_worker.models import Element
from arkindex_worker.worker import ElementsWorker

# Initialize the logger to provide feedback about the worker's execution to the final user
logging.basicConfig(
    format="%(asctime)s %(levelname)s/%(name)s: %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


# Create a worker inheriting from the `ElementsWorker` class provided by the `arkindex-base-worker` package
class BasicWorker(ElementsWorker):
    def process_element(self, element: Element) -> None:
        """Process a single Arkindex element at once and publish a simple transcription on it.

        :param Element element: The element currently being processed from the element list
        """
        try:
            # Create the "Hello world!" transcription on the current element
            # Helper: `TranscriptionMixin.create_transcription` from the `arkindex-base-worker` package
            transcription = self.create_transcription(
                element=element,
                text="Hello world!",
                confidence=1.0,
            )

            # Output feedback when a transcription is successfully created
            logger.info(
                f"A transcription with the ID {transcription['id']} was successfully created on element {element.id}."
            )

        except Exception as e:
            # Output feedback when failing to create a transcription
            logger.error(
                f"Failed to create a transcription on element {element.id}: {e}"
            )


def main() -> None:
    BasicWorker(
        description="Tooled Python worker to create a transcription on Arkindex elements"
    ).run()


if __name__ == "__main__":
    main()



from uuid import UUID

import pytest

from arkindex_worker.cache import (
    CachedElement,
)
from arkindex_worker.models import Element

from . import BASE_API_CALLS


def test_list_element_parents_wrong_element(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.list_element_parents(element=None)

    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.list_element_parents(element="not element type")


def test_list_element_parents_wrong_folder(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="folder should be of type bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            folder="not bool",
        )


def test_list_element_parents_wrong_name(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="name should be of type str"):
        mock_elements_worker.list_element_parents(
            element=elt,
            name=1234,
        )


def test_list_element_parents_wrong_recursive(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="recursive should be of type bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            recursive="not bool",
        )


def test_list_element_parents_wrong_type(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="type should be of type str"):
        mock_elements_worker.list_element_parents(
            element=elt,
            type=1234,
        )


def test_list_element_parents_wrong_with_classes(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="with_classes should be of type bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            with_classes="not bool",
        )


def test_list_element_parents_wrong_with_corpus(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="with_corpus should be of type bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            with_corpus="not bool",
        )


def test_list_element_parents_wrong_with_has_children(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="with_has_children should be of type bool"
    ):
        mock_elements_worker.list_element_parents(
            element=elt,
            with_has_children="not bool",
        )


def test_list_element_parents_wrong_with_zone(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="with_zone should be of type bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            with_zone="not bool",
        )


def test_list_element_parents_wrong_with_metadata(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match="with_metadata should be of type bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            with_metadata="not bool",
        )


@pytest.mark.parametrize(
    ("param", "value"),
    [
        ("worker_run", 1234),
        ("transcription_worker_run", 1234),
    ],
)
def test_list_element_parents_wrong_worker_run(mock_elements_worker, param, value):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(AssertionError, match=f"{param} should be of type str or bool"):
        mock_elements_worker.list_element_parents(
            element=elt,
            **{param: value},
        )


@pytest.mark.parametrize(
    ("param", "alternative", "value"),
    [
        ("worker_version", "worker_run", 1234),
        ("transcription_worker_version", "transcription_worker_run", 1234),
    ],
)
def test_list_element_parents_wrong_worker_version(
    mock_elements_worker, param, alternative, value
):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    # WARNING: pytest.deprecated_call must be placed BEFORE pytest.raises, otherwise `match` argument won't be checked
    with (
        pytest.deprecated_call(
            match=f"`{param}` usage is deprecated. Consider using `{alternative}` instead."
        ),
        pytest.raises(AssertionError, match=f"{param} should be of type str or bool"),
    ):
        mock_elements_worker.list_element_parents(
            element=elt,
            **{param: value},
        )


@pytest.mark.parametrize(
    "param",
    [
        "worker_run",
        "transcription_worker_run",
    ],
)
def test_list_element_parents_wrong_bool_worker_run(mock_elements_worker, param):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match=f"if of type bool, {param} can only be set to False"
    ):
        mock_elements_worker.list_element_parents(
            element=elt,
            **{param: True},
        )


@pytest.mark.parametrize(
    ("param", "alternative"),
    [
        ("worker_version", "worker_run"),
        ("transcription_worker_version", "transcription_worker_run"),
    ],
)
def test_list_element_parents_wrong_bool_worker_version(
    mock_elements_worker, param, alternative
):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    # WARNING: pytest.deprecated_call must be placed BEFORE pytest.raises, otherwise `match` argument won't be checked
    with (
        pytest.deprecated_call(
            match=f"`{param}` usage is deprecated. Consider using `{alternative}` instead."
        ),
        pytest.raises(
            AssertionError, match=f"if of type bool, {param} can only be set to False"
        ),
    ):
        mock_elements_worker.list_element_parents(
            element=elt,
            **{param: True},
        )


def test_list_element_parents_api_error(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.GET,
        "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        status=418,
    )

    with pytest.raises(
        Exception, match="Stopping pagination as data will be incomplete"
    ):
        next(mock_elements_worker.list_element_parents(element=elt))

    assert len(responses.calls) == len(BASE_API_CALLS) + 5
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        # We do 5 retries
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        ),
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        ),
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        ),
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        ),
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        ),
    ]


def test_list_element_parents(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    expected_parents = [
        {
            "id": "0000",
            "type": "page",
            "name": "Test",
            "corpus": {},
            "thumbnail_url": None,
            "zone": {},
            "best_classes": None,
            "has_children": None,
            "worker_version_id": None,
            "worker_run_id": None,
        },
        {
            "id": "1111",
            "type": "page",
            "name": "Test 2",
            "corpus": {},
            "thumbnail_url": None,
            "zone": {},
            "best_classes": None,
            "has_children": None,
            "worker_version_id": None,
            "worker_run_id": None,
        },
        {
            "id": "2222",
            "type": "page",
            "name": "Test 3",
            "corpus": {},
            "thumbnail_url": None,
            "zone": {},
            "best_classes": None,
            "has_children": None,
            "worker_version_id": None,
            "worker_run_id": None,
        },
    ]
    responses.add(
        responses.GET,
        "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        status=200,
        json={
            "count": 3,
            "next": None,
            "results": expected_parents,
        },
    )

    for idx, parent in enumerate(
        mock_elements_worker.list_element_parents(element=elt)
    ):
        assert parent == expected_parents[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/",
        ),
    ]


def test_list_element_parents_manual_worker_version(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    expected_parents = [
        {
            "id": "0000",
            "type": "page",
            "name": "Test",
            "corpus": {},
            "thumbnail_url": None,
            "zone": {},
            "best_classes": None,
            "has_children": None,
            "worker_version_id": None,
            "worker_run_id": None,
        }
    ]
    responses.add(
        responses.GET,
        "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/?worker_version=False",
        status=200,
        json={
            "count": 1,
            "next": None,
            "results": expected_parents,
        },
    )

    with pytest.deprecated_call(
        match="`worker_version` usage is deprecated. Consider using `worker_run` instead."
    ):
        for idx, parent in enumerate(
            mock_elements_worker.list_element_parents(element=elt, worker_version=False)
        ):
            assert parent == expected_parents[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/?worker_version=False",
        ),
    ]


def test_list_element_parents_manual_worker_run(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    expected_parents = [
        {
            "id": "0000",
            "type": "page",
            "name": "Test",
            "corpus": {},
            "thumbnail_url": None,
            "zone": {},
            "best_classes": None,
            "has_children": None,
            "worker_version_id": None,
            "worker_run_id": None,
        }
    ]
    responses.add(
        responses.GET,
        "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/?worker_run=False",
        status=200,
        json={
            "count": 1,
            "next": None,
            "results": expected_parents,
        },
    )

    for idx, parent in enumerate(
        mock_elements_worker.list_element_parents(element=elt, worker_run=False)
    ):
        assert parent == expected_parents[idx]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/elements/12341234-1234-1234-1234-123412341234/parents/?worker_run=False",
        ),
    ]


def test_list_element_parents_with_cache_unhandled_param(
    mock_elements_worker_with_cache,
):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError,
        match="When using the local cache, you can only filter by 'type' and/or 'worker_version' and/or 'worker_run'",
    ):
        mock_elements_worker_with_cache.list_element_parents(
            element=elt, with_corpus=True
        )


@pytest.mark.usefixtures("_mock_cached_elements")
@pytest.mark.parametrize(
    ("filters", "expected_id"),
    [
        # Filter on element
        (
            {
                "element": CachedElement(id="11111111-1111-1111-1111-111111111111"),
            },
            "12341234-1234-1234-1234-123412341234",
        ),
        # Filter on element and double_page
        (
            {
                "element": CachedElement(id="22222222-2222-2222-2222-222222222222"),
                "type": "double_page",
            },
            "12341234-1234-1234-1234-123412341234",
        ),
        # Filter on element and worker run
        (
            {
                "element": CachedElement(id="22222222-2222-2222-2222-222222222222"),
                "worker_run": "56785678-5678-5678-5678-567856785678",
            },
            "12341234-1234-1234-1234-123412341234",
        ),
        # Filter on element, manual worker run
        (
            {
                "element": CachedElement(id="12341234-1234-1234-1234-123412341234"),
                "worker_run": False,
            },
            "99999999-9999-9999-9999-999999999999",
        ),
    ],
)
def test_list_element_parents_with_cache(
    responses,
    mock_elements_worker_with_cache,
    filters,
    expected_id,
):
    # Check we have 5 elements already present in database
    assert CachedElement.select().count() == 5

    # Query database through cache
    elements = mock_elements_worker_with_cache.list_element_parents(**filters)
    assert elements.count() == 1
    for parent in elements.order_by("id"):
        assert parent.id == UUID(expected_id)

    # Check the worker never hits the API for elements
    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS


@pytest.mark.usefixtures("_mock_cached_elements")
@pytest.mark.parametrize(
    ("filters", "expected_id"),
    [
        # Filter on element and worker version
        (
            {
                "element": CachedElement(id="33333333-3333-3333-3333-333333333333"),
                "worker_version": "56785678-5678-5678-5678-567856785678",
            },
            "12341234-1234-1234-1234-123412341234",
        ),
        # Filter on element, type double_page and worker version
        (
            {
                "element": CachedElement(id="11111111-1111-1111-1111-111111111111"),
                "type": "double_page",
                "worker_version": "56785678-5678-5678-5678-567856785678",
            },
            "12341234-1234-1234-1234-123412341234",
        ),
        # Filter on element, manual worker version
        (
            {
                "element": CachedElement(id="12341234-1234-1234-1234-123412341234"),
                "worker_version": False,
            },
            "99999999-9999-9999-9999-999999999999",
        ),
    ],
)
def test_list_element_parents_with_cache_deprecation(
    responses,
    mock_elements_worker_with_cache,
    filters,
    expected_id,
):
    # Check we have 5 elements already present in database
    assert CachedElement.select().count() == 5

    with pytest.deprecated_call(
        match="`worker_version` usage is deprecated. Consider using `worker_run` instead."
    ):
        # Query database through cache
        elements = mock_elements_worker_with_cache.list_element_parents(**filters)
    assert elements.count() == 1
    for parent in elements.order_by("id"):
        assert parent.id == UUID(expected_id)

    # Check the worker never hits the API for elements
    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS



import json

import pytest

from arkindex.exceptions import ErrorResponse

from . import BASE_API_CALLS


@pytest.mark.parametrize(
    ("data", "error_message"),
    [
        (None, "url shouldn't be null and should be of type str"),
        (1234, "url shouldn't be null and should be of type str"),
    ],
)
def test_create_iiif_url_wrong_data(data, error_message, mock_elements_worker):
    with pytest.raises(AssertionError, match=error_message):
        mock_elements_worker.create_iiif_url(url=data)


def test_create_iiif_url_api_error(responses, mock_elements_worker):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/image/iiif/url/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_iiif_url("http://url/to/my/image")

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("POST", "http://testserver/api/v1/image/iiif/url/")]


def test_create_iiif_url(responses, mock_elements_worker):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/image/iiif/url/",
        status=201,
        json={
            "id": "cafecafe-cafe-cafe-cafe-cafecafecafe",
            "url": "http://url/to/my/image",
            "status": "checked",
            "server": {
                "id": 5,
                "display_name": "My server",
                "url": "http://url/to/my",
                "max_width": 42,
                "max_height": 42,
            },
        },
    )

    image = mock_elements_worker.create_iiif_url("http://url/to/my/image")

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("POST", "http://testserver/api/v1/image/iiif/url/")]
    assert json.loads(responses.calls[-1].request.body) == {
        "url": "http://url/to/my/image"
    }
    assert image.id == "cafecafe-cafe-cafe-cafe-cafecafecafe"



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent
FIXTURES_DIR = BASE_DIR / "data"
SAMPLES_DIR = BASE_DIR / "samples"

CORPUS_ID = "11111111-1111-1111-1111-111111111111"
PROCESS_ID = "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff"



import importlib


def test_dummy():
    assert True


def test_import():
    """Import our newly created module, through importlib to avoid parsing issues"""
    worker = importlib.import_module("worker_demo.worker")
    assert hasattr(worker, "Demo")
    assert hasattr(worker.Demo, "process_element")



import json
import logging

import pytest

from arkindex.exceptions import ErrorResponse
from arkindex_worker.models import Dataset, Element, Set
from arkindex_worker.worker.dataset import DatasetState
from tests import PROCESS_ID
from tests.test_elements_worker import BASE_API_CALLS


def test_list_process_sets_readonly_error(mock_dataset_worker):
    # Set worker in read_only mode
    mock_dataset_worker.worker_run_id = None
    assert mock_dataset_worker.is_read_only

    with pytest.raises(
        AssertionError, match="This helper is not available in read-only mode."
    ):
        mock_dataset_worker.list_process_sets()


def test_list_process_sets_api_error(responses, mock_dataset_worker):
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/process/{PROCESS_ID}/sets/",
        status=418,
    )

    with pytest.raises(
        Exception, match="Stopping pagination as data will be incomplete"
    ):
        next(mock_dataset_worker.list_process_sets())

    assert len(responses.calls) == len(BASE_API_CALLS) + 5
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        # The API call is retried 5 times
        ("GET", f"http://testserver/api/v1/process/{PROCESS_ID}/sets/"),
        ("GET", f"http://testserver/api/v1/process/{PROCESS_ID}/sets/"),
        ("GET", f"http://testserver/api/v1/process/{PROCESS_ID}/sets/"),
        ("GET", f"http://testserver/api/v1/process/{PROCESS_ID}/sets/"),
        ("GET", f"http://testserver/api/v1/process/{PROCESS_ID}/sets/"),
    ]


def test_list_process_sets(
    responses,
    mock_dataset_worker,
):
    expected_results = [
        {
            "id": "set_1",
            "dataset": {
                "id": "dataset_1",
                "name": "Dataset 1",
                "description": "My first great dataset",
                "sets": [
                    {"id": "set_1", "name": "train"},
                    {"id": "set_2", "name": "val"},
                ],
                "state": "open",
                "corpus_id": "corpus_id",
                "creator": "test@teklia.com",
                "task_id": "task_id_1",
            },
            "set_name": "train",
        },
        {
            "id": "set_2",
            "dataset": {
                "id": "dataset_1",
                "name": "Dataset 1",
                "description": "My first great dataset",
                "sets": [
                    {"id": "set_1", "name": "train"},
                    {"id": "set_2", "name": "val"},
                ],
                "state": "open",
                "corpus_id": "corpus_id",
                "creator": "test@teklia.com",
                "task_id": "task_id_1",
            },
            "set_name": "val",
        },
        {
            "id": "set_3",
            "dataset": {
                "id": "dataset_2",
                "name": "Dataset 2",
                "description": "My second great dataset",
                "sets": [{"id": "set_3", "name": "my_set"}],
                "state": "complete",
                "corpus_id": "corpus_id",
                "creator": "test@teklia.com",
                "task_id": "task_id_2",
            },
            "set_name": "my_set",
        },
    ]
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/process/{PROCESS_ID}/sets/",
        status=200,
        json={
            "count": 3,
            "next": None,
            "results": expected_results,
        },
    )

    for idx, dataset_set in enumerate(mock_dataset_worker.list_process_sets()):
        assert isinstance(dataset_set, Set)
        assert dataset_set.name == expected_results[idx]["set_name"]

        assert isinstance(dataset_set.dataset, Dataset)
        assert dataset_set.dataset == expected_results[idx]["dataset"]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", f"http://testserver/api/v1/process/{PROCESS_ID}/sets/"),
    ]


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Set
        (
            {"dataset_set": None},
            "dataset_set shouldn't be null and should be a Set",
        ),
        (
            {"dataset_set": "not Set type"},
            "dataset_set shouldn't be null and should be a Set",
        ),
    ],
)
def test_list_set_elements_wrong_param_dataset_set(mock_dataset_worker, payload, error):
    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.list_set_elements(**payload)


def test_list_set_elements_api_error(
    responses, mock_dataset_worker, default_dataset, default_train_set
):
    query_params = f"?set={default_train_set.name}&with_count=true"
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        status=418,
    )

    with pytest.raises(
        Exception, match="Stopping pagination as data will be incomplete"
    ):
        next(mock_dataset_worker.list_set_elements(dataset_set=default_train_set))

    assert len(responses.calls) == len(BASE_API_CALLS) + 5
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        # The API call is retried 5 times
        (
            "GET",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        ),
        (
            "GET",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        ),
        (
            "GET",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        ),
        (
            "GET",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        ),
        (
            "GET",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        ),
    ]


def test_list_set_elements(
    responses,
    mock_dataset_worker,
    default_dataset,
    default_train_set,
):
    expected_results = [
        {
            "set": "train",
            "element": {
                "id": "element_1",
                "type": "page",
                "name": "1",
                "corpus": {},
                "thumbnail_url": None,
                "zone": {},
                "best_classes": None,
                "has_children": None,
                "worker_version_id": None,
                "worker_run_id": None,
            },
        }
    ]
    expected_results.append({**expected_results[-1]})
    expected_results[-1]["element"]["id"] = "element_2"
    expected_results[-1]["element"]["name"] = "2"

    query_params = f"?set={default_train_set.name}&with_count=true"
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        status=200,
        json={
            "count": 2,
            "next": None,
            "results": expected_results,
        },
    )

    for idx, element in enumerate(
        mock_dataset_worker.list_set_elements(dataset_set=default_train_set)
    ):
        assert isinstance(element, Element)
        assert element == expected_results[idx]["element"]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/elements/{query_params}",
        )
    ]


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # Dataset
        (
            {"dataset": None},
            "dataset shouldn't be null and should be a Dataset",
        ),
        (
            {"dataset": "not dataset type"},
            "dataset shouldn't be null and should be a Dataset",
        ),
    ],
)
def test_update_dataset_state_wrong_param_dataset(
    mock_dataset_worker, default_dataset, payload, error
):
    api_payload = {
        "dataset": Dataset(**default_dataset),
        "state": DatasetState.Building,
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.update_dataset_state(**api_payload)


@pytest.mark.parametrize(
    ("payload", "error"),
    [
        # DatasetState
        (
            {"state": None},
            "state shouldn't be null and should be a str from DatasetState",
        ),
        (
            {"state": "not dataset type"},
            "state shouldn't be null and should be a str from DatasetState",
        ),
    ],
)
def test_update_dataset_state_wrong_param_state(
    mock_dataset_worker, default_dataset, payload, error
):
    api_payload = {
        "dataset": Dataset(**default_dataset),
        "state": DatasetState.Building,
        **payload,
    }

    with pytest.raises(AssertionError, match=error):
        mock_dataset_worker.update_dataset_state(**api_payload)


def test_update_dataset_state_readonly_error(
    caplog, mock_dev_dataset_worker, default_dataset
):
    api_payload = {
        "dataset": Dataset(**default_dataset),
        "state": DatasetState.Building,
    }

    assert not mock_dev_dataset_worker.update_dataset_state(**api_payload)
    assert [(level, message) for _, level, message in caplog.record_tuples] == [
        (
            logging.WARNING,
            "Cannot update dataset as this worker is in read-only mode",
        ),
    ]


def test_update_dataset_state_api_error(
    responses, mock_dataset_worker, default_dataset
):
    responses.add(
        responses.PATCH,
        f"http://testserver/api/v1/datasets/{default_dataset.id}/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_dataset_worker.update_dataset_state(
            dataset=default_dataset,
            state=DatasetState.Building,
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("PATCH", f"http://testserver/api/v1/datasets/{default_dataset.id}/")
    ]


def test_update_dataset_state(
    responses,
    mock_dataset_worker,
    default_dataset,
):
    dataset_response = {
        "name": "My dataset",
        "description": "A super dataset built by me",
        "sets": ["set_1", "set_2", "set_3"],
        "state": DatasetState.Building.value,
    }
    responses.add(
        responses.PATCH,
        f"http://testserver/api/v1/datasets/{default_dataset.id}/",
        status=200,
        json=dataset_response,
    )

    updated_dataset = mock_dataset_worker.update_dataset_state(
        dataset=default_dataset,
        state=DatasetState.Building,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "PATCH",
            f"http://testserver/api/v1/datasets/{default_dataset.id}/",
        ),
    ]
    assert json.loads(responses.calls[-1].request.body) == {
        "state": DatasetState.Building.value
    }
    assert updated_dataset == Dataset(**{**default_dataset, **dataset_response})



import pytest
from requests import HTTPError

from arkindex_worker.cache import CachedElement
from arkindex_worker.models import Element


@pytest.mark.parametrize(
    ("zone", "expected_url"),
    [
        (None, None),
        (
            {"image": {"url": "http://something/", "server": {"version": 2}}},
            "http://something/full/full/0/default.jpg",
        ),
        (
            {"image": {"url": "http://something", "server": {"version": 2}}},
            "http://something/full/full/0/default.jpg",
        ),
        (
            {"image": {"url": "http://something/", "server": {"version": 3}}},
            "http://something/full/max/0/default.jpg",
        ),
    ],
)
def test_image_url(zone, expected_url):
    url = Element({"zone": None}).image_url()
    assert not url


def test_image_url_iiif_resize():
    url = Element(
        {"zone": {"image": {"url": "http://something/", "server": {"version": 2}}}}
    ).image_url(500)
    assert url == "http://something/full/500/0/default.jpg"


def test_image_url_s3():
    url = Element(
        {
            "zone": {
                "image": {"s3_url": "http://s3/something", "url": "http://something/"}
            }
        }
    ).image_url()
    assert url == "http://s3/something"


def test_image_url_s3_resize():
    url = Element(
        {
            "zone": {
                "image": {"s3_url": "http://s3/something", "url": "http://something/"}
            }
        }
    ).image_url(500)
    assert url == "http://s3/something"


def test_open_image(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [181, 0], [181, 240], [0, 240], [0, 0]],
            },
            "rotation_angle": 0,
            "mirrored": False,
        },
    )
    assert elt.open_image(use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_resize_portrait(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "width": 400,
                    "height": 600,
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [400, 0], [400, 600], [0, 600], [0, 0]],
            },
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    # Resize = original size
    assert elt.open_image(max_height=600, use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )
    # Resize = smaller height
    assert elt.open_image(max_height=400, use_full_image=True) == "an image!"
    assert open_mock.call_count == 2
    assert open_mock.call_args == mocker.call(
        "http://something/full/,400/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )
    # Resize = bigger height
    assert elt.open_image(max_height=800, use_full_image=True) == "an image!"
    assert open_mock.call_count == 3
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_resize_partial_element(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "width": 400,
                    "height": 600,
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [200, 0], [200, 600], [0, 600], [0, 0]],
            },
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    assert elt.open_image(max_height=400, use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/,400/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_resize_landscape(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "width": 600,
                    "height": 400,
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [600, 0], [600, 400], [0, 400], [0, 0]],
            },
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    # Resize = original size
    assert elt.open_image(max_width=600, use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )
    # Resize = smaller width
    assert elt.open_image(max_width=400, use_full_image=True) == "an image!"
    assert open_mock.call_count == 2
    assert open_mock.call_args == mocker.call(
        "http://something/full/400,/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )
    # Resize = bigger width
    assert elt.open_image(max_width=800, use_full_image=True) == "an image!"
    assert open_mock.call_count == 3
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_resize_square(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "width": 400,
                    "height": 400,
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [400, 0], [400, 400], [0, 400], [0, 0]],
            },
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    # Resize = original size
    assert (
        elt.open_image(
            max_width=400,
            max_height=400,
            use_full_image=True,
        )
        == "an image!"
    )
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )
    # Resize = smaller
    assert (
        elt.open_image(
            max_width=200,
            max_height=200,
            use_full_image=True,
        )
        == "an image!"
    )
    assert open_mock.call_count == 2
    assert open_mock.call_args == mocker.call(
        "http://something/full/200,200/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )
    # Resize = bigger
    assert (
        elt.open_image(
            max_width=800,
            max_height=800,
            use_full_image=True,
        )
        == "an image!"
    )
    assert open_mock.call_count == 3
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_resize_tiles(mocker):
    mocker.patch("arkindex_worker.image.open_image", return_value="an image!")
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "server": {"max_width": 600, "max_height": 600},
                },
                "polygon": [[0, 0], [800, 0], [800, 800], [0, 800], [0, 0]],
            }
        }
    )
    with pytest.raises(NotImplementedError):
        elt.open_image(max_width=400)


def test_open_image_requires_zone():
    with pytest.raises(ValueError, match="Element 42 has no zone"):
        Element({"id": "42"}).open_image()


def test_open_image_s3(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {"image": {"url": "http://something", "s3_url": "http://s3url"}},
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    assert elt.open_image(use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://s3url", rotation_angle=0, mirrored=False
    )


def test_open_image_s3_retry(mocker):
    response_mock = mocker.MagicMock()
    response_mock.status_code = 403
    mocker.patch(
        "arkindex_worker.image.open_image",
        return_value="an image!",
        side_effect=HTTPError(response=response_mock),
    )

    elt = Element(
        {
            "id": "cafe",
            "zone": {"image": {"url": "http://something", "s3_url": "http://oldurl"}},
            "rotation_angle": 0,
            "mirrored": False,
        }
    )

    with pytest.raises(NotImplementedError):
        elt.open_image(use_full_image=True)


def test_open_image_s3_retry_once(mocker):
    response_mock = mocker.MagicMock()
    response_mock.status_code = 403
    mocker.patch(
        "arkindex_worker.image.open_image",
        side_effect=HTTPError(response=response_mock),
    )
    elt = Element(
        {
            "id": "cafe",
            "zone": {"image": {"url": "http://something", "s3_url": "http://oldurl"}},
            "rotation_angle": 0,
            "mirrored": False,
        }
    )

    with pytest.raises(NotImplementedError):
        elt.open_image(use_full_image=True)


def test_open_image_use_full_image_false(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {"url": "http://something", "s3_url": "http://s3url"},
                "url": "http://zoneurl/0,0,400,600/full/0/default.jpg",
            },
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    assert elt.open_image(use_full_image=False) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://zoneurl/0,0,400,600/full/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_resize_use_full_image_false(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "width": 400,
                    "height": 600,
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [400, 0], [400, 600], [0, 600], [0, 0]],
                "url": "http://zoneurl/0,0,400,600/full/0/default.jpg",
            },
            "rotation_angle": 0,
            "mirrored": False,
        }
    )
    # Resize = smaller
    assert elt.open_image(max_height=200, use_full_image=False) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://zoneurl/0,0,400,600/,200/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_open_image_rotation_mirror(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "server": {"max_width": None, "max_height": None},
                },
                "polygon": [[0, 0], [181, 0], [181, 240], [0, 240], [0, 0]],
            },
            "rotation_angle": 42,
            "mirrored": True,
        },
    )
    assert elt.open_image(use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/full/0/default.jpg",
        rotation_angle=42,
        mirrored=True,
    )


def test_open_image_iiif_3(mocker):
    open_mock = mocker.patch(
        "arkindex_worker.image.open_image", return_value="an image!"
    )
    elt = Element(
        {
            "zone": {
                "image": {
                    "url": "http://something",
                    "server": {
                        "max_width": None,
                        "max_height": None,
                        "version": 3,
                    },
                },
                "polygon": [[0, 0], [181, 0], [181, 240], [0, 240], [0, 0]],
            },
            "rotation_angle": 0,
            "mirrored": False,
        },
    )
    assert elt.open_image(use_full_image=True) == "an image!"
    assert open_mock.call_count == 1
    assert open_mock.call_args == mocker.call(
        "http://something/full/max/0/default.jpg",
        rotation_angle=0,
        mirrored=False,
    )


def test_setattr_setitem():
    element = Element({"name": "something"})
    element.type = "page"
    assert dict(element) == {"name": "something", "type": "page"}


def test_element_polygon():
    polygon = [[0, 0], [181, 0], [181, 240], [0, 240], [0, 0]]
    element = Element({"zone": {"polygon": polygon}})
    cached_element = CachedElement(polygon=polygon)
    assert element.polygon == polygon
    assert element.polygon == cached_element.polygon


def test_element_no_polygon():
    element = Element(id="element_id")
    with pytest.raises(ValueError, match="Element element_id has no zone"):
        _ = element.polygon



import logging
import math
import uuid
from io import BytesIO
from operator import attrgetter
from pathlib import Path

import pytest
from PIL import Image, ImageChops, ImageOps
from requests import HTTPError

import arkindex_worker.image
from arkindex_worker.cache import CachedElement, create_tables, init_cache_db
from arkindex_worker.image import (
    IIIF_FULL,
    IIIF_MAX,
    BoundingBox,
    download_image,
    download_tiles,
    open_image,
    polygon_bounding_box,
    resized_images,
    revert_orientation,
    trim_polygon,
    update_pillow_image_size_limit,
    upload_image,
)
from arkindex_worker.models import Element
from tests import FIXTURES_DIR

TILE = FIXTURES_DIR / "test_image.jpg"
FULL_IMAGE = FIXTURES_DIR / "tiled_image.jpg"
ROTATED_IMAGE = FIXTURES_DIR / "rotated_image.jpg"
MIRRORED_IMAGE = FIXTURES_DIR / "mirrored_image.jpg"
ROTATED_MIRRORED_IMAGE = FIXTURES_DIR / "rotated_mirrored_image.jpg"
TEST_IMAGE = {"width": 800, "height": 300}


@pytest.fixture
def mock_page():
    class Page(Element):
        @property
        def crop(self):
            # Image from Socface (https://socface.site.ined.fr/) project (AD026)
            image = Image.open(FIXTURES_DIR / "AD026_6M_00505_0001_0373.jpg")
            x, y, element_width, element_height = polygon_bounding_box(
                self.zone.polygon
            )
            return image.crop(box=(x, y, x + element_width, y + element_height))

        def open_image(
            self,
            *args,
            max_width: int | None = None,
            max_height: int | None = None,
            use_full_image: bool | None = False,
            **kwargs,
        ) -> Image.Image:
            crop = self.crop.copy()
            crop.thumbnail(
                size=(
                    max_width or self.zone.image.width,
                    max_height or self.zone.image.height,
                )
            )
            return crop

    return Page(
        id="page_id",
        name="1",
        zone={
            "polygon": [[0, 0], [2000, 0], [2000, 3000], [0, 3000], [0, 0]],
            "image": {"width": 2000, "height": 3000},
        },
        rotation_angle=0,
        mirrored=False,
    )


def _root_mean_square(img_a, img_b):
    """
    Get the root-mean-square difference between two images for fuzzy matching
    See https://effbot.org/zone/pil-comparing-images.htm
    """
    h = ImageChops.difference(img_a, img_b).histogram()
    return math.sqrt(
        sum((value * ((idx % 256) ** 2) for idx, value in enumerate(h)))
        / float(img_a.size[0] * img_a.size[1])
    )


@pytest.mark.parametrize(
    ("max_image_pixels", "expected_image_pixels"),
    [
        # Pillow Image size limit not updated
        (None, Image.MAX_IMAGE_PIXELS),
        # Pillow Image size limit set to None
        ("0", None),
        (0, None),
        # Update Pillow Image size limit
        ("1", 1),
        (1, 1),
    ],
)
def test_update_pillow_image_size_limit(max_image_pixels, expected_image_pixels):
    MAX_IMAGE_PIXELS = Image.MAX_IMAGE_PIXELS

    @update_pillow_image_size_limit
    def function() -> int | None:
        return Image.MAX_IMAGE_PIXELS

    assert function(max_image_pixels=max_image_pixels) == expected_image_pixels
    assert Image.MAX_IMAGE_PIXELS == MAX_IMAGE_PIXELS


@pytest.mark.parametrize(
    ("id_key", "resize"),
    [
        # IIIF version 2
        ("@id", "full"),
        # IIIF version 3
        ("id", "max"),
    ],
)
def test_download_tiles(responses, id_key, resize):
    expected = Image.open(FULL_IMAGE).convert("RGB")
    tile_bytes = TILE.read_bytes()

    responses.add(
        responses.GET,
        "http://nowhere/info.json",
        json={
            id_key: "http://nowhere",
            "width": 543,
            "height": 720,
            "tiles": [
                {"width": 181, "height": 240},
            ],
        },
    )

    for x in (0, 181, 362):
        for y in (0, 240, 480):
            responses.add(
                responses.GET,
                f"http://nowhere/{x},{y},181,240/{resize}/0/default.jpg",
                body=tile_bytes,
            )

    actual = download_tiles("http://nowhere")

    assert _root_mean_square(expected, actual) <= 5.0


def test_download_tiles_crop(responses):
    """
    Ensure download_tiles does not care about tiles that are slightly bigger than expected
    (occasional issue with the Harvard IDS image server where 1024×1024 tiles sometimes are returned as 1024x1025)
    """
    expected = Image.open(FULL_IMAGE).convert("RGB")
    tile_bytes = BytesIO()
    # Add one extra pixel to each tile to return slightly bigger tiles
    ImageOps.pad(Image.open(TILE), (181, 241)).save(tile_bytes, format="jpeg")

    tile_bytes = tile_bytes.getvalue()

    responses.add(
        responses.GET,
        "http://nowhere/info.json",
        json={"width": 543, "height": 720, "tiles": [{"width": 181, "height": 240}]},
    )

    for x in (0, 181, 362):
        for y in (0, 240, 480):
            responses.add(
                responses.GET,
                f"http://nowhere/{x},{y},181,240/full/0/default.jpg",
                body=tile_bytes,
            )

    actual = download_tiles("http://nowhere")

    assert _root_mean_square(expected, actual) <= 5.0


def test_download_tiles_small(responses):
    small_tile = BytesIO()
    Image.new("RGB", (1, 1)).save(small_tile, format="jpeg")
    small_tile.seek(0)

    responses.add(
        responses.GET,
        "http://nowhere/info.json",
        json={"width": 543, "height": 720, "tiles": [{"width": 181, "height": 240}]},
    )

    responses.add(
        responses.GET,
        "http://nowhere/0,0,181,240/full/0/default.jpg",
        body=small_tile.read(),
    )

    with pytest.raises(
        ValueError, match="Expected size 181×240 for tile 0,0, but got 1×1"
    ):
        download_tiles("http://nowhere")


@pytest.mark.parametrize(
    ("path", "is_local"),
    [
        ("http://somewhere/test.jpg", False),
        ("https://somewhere/test.jpg", False),
        ("path/to/something", True),
        ("/absolute/path/to/something", True),
    ],
)
def test_open_image(path, is_local, monkeypatch):
    """Check if the path triggers a local load or a remote one"""

    monkeypatch.setattr(Path, "exists", lambda x: True)
    monkeypatch.setattr(Image, "open", lambda x: Image.new("RGB", (1, 10)))
    monkeypatch.setattr(
        "arkindex_worker.image.download_image", lambda x: Image.new("RGB", (10, 1))
    )
    image = open_image(path)

    if is_local:
        assert image.size == (1, 10)
    else:
        assert image.size == (10, 1)


@pytest.mark.parametrize(
    ("rotation_angle", "mirrored", "expected_path"),
    [
        (0, False, TILE),
        (45, False, ROTATED_IMAGE),
        (0, True, MIRRORED_IMAGE),
        (45, True, ROTATED_MIRRORED_IMAGE),
    ],
)
def test_open_image_rotate_mirror(rotation_angle, mirrored, expected_path):
    expected = Image.open(expected_path).convert("RGB")
    actual = open_image(str(TILE), rotation_angle=rotation_angle, mirrored=mirrored)
    actual.save(f"/tmp/{rotation_angle}_{mirrored}.jpg")
    assert _root_mean_square(expected, actual) <= 15.0


@pytest.mark.parametrize(
    ("polygon", "error"),
    [
        # Polygon isn't a list or tuple
        (
            {
                "polygon": [
                    [99, 200],
                    [25, 224],
                    [0, 0],
                    [0, 300],
                    [102, 300],
                    [260, 300],
                    [288, 295],
                    [296, 260],
                    [352, 259],
                    [106, 210],
                    [197, 206],
                    [99, 208],
                ]
            },
            "Polygon must be a valid list or tuple of points.",
        ),
        # Polygon hasn't enough points
        (
            [
                [99, 200],
                [25, 224],
            ],
            "Polygon should have at least three points.",
        ),
        # Point coordinates are not integers
        (
            [
                [9.9, 200],
                [25, 224],
                [0, 0],
                [0, 300],
                [102, 300],
                [260, 300],
                [288, 295],
                [296, 260],
                [352, 259],
                [106, 210],
                [197, 206],
                [99, 20.8],
            ],
            "Polygon point coordinates must be integers.",
        ),
        # Point coordinates are not lists or tuples
        (
            [
                [12, 56],
                [29, 60],
                [35, 61],
                "[42, 59]",
                [58, 57],
                [65, 61],
                [72, 57],
                [12, 56],
            ],
            "Polygon points must be tuples or lists.",
        ),
        # Point coordinates are not lists or tuples of length 2
        (
            [
                [12, 56],
                [29, 60, 3],
                [35, 61],
                [42, 59],
                [58, 57],
                [65, 61],
                [72, 57],
                [12, 56],
            ],
            "Polygon points must be tuples or lists of 2 elements.",
        ),
        # None of the polygon's points are inside the image
        (
            [
                [999, 200],
                [1097, 224],
                [1020, 251],
                [1232, 350],
                [1312, 364],
                [1325, 310],
                [1318, 295],
                [1296, 260],
                [1352, 259],
                [1006, 210],
                [997, 206],
                [999, 200],
            ],
            "This polygon is entirely outside the image's bounds.",
        ),
    ],
)
def test_trim_polygon_errors(polygon, error):
    with pytest.raises(AssertionError, match=error):
        trim_polygon(polygon, TEST_IMAGE["width"], TEST_IMAGE["height"])


def test_trim_polygon_negative_coordinates():
    """
    Negative coordinates are ignored and replaced by 0 with no error being thrown
    """
    polygon = [
        [99, 200],
        [25, 224],
        [-8, -52],
        [-12, 350],
        [102, 364],
        [260, 310],
        [288, 295],
        [296, 260],
        [352, 259],
        [106, 210],
        [197, 206],
        [99, 200],
    ]
    expected_polygon = [
        [99, 200],
        [25, 224],
        [0, 0],
        [0, 300],
        [102, 300],
        [260, 300],
        [288, 295],
        [296, 260],
        [352, 259],
        [106, 210],
        [197, 206],
        [99, 200],
    ]
    assert (
        trim_polygon(polygon, TEST_IMAGE["width"], TEST_IMAGE["height"])
        == expected_polygon
    )


def test_trim_polygon_partially_outside_image():
    polygon = [
        [99, 200],
        [197, 224],
        [120, 251],
        [232, 350],
        [312, 364],
        [325, 310],
        [318, 295],
        [296, 260],
        [352, 259],
        [106, 210],
        [197, 206],
        [99, 200],
    ]
    expected_polygon = [
        [99, 200],
        [197, 224],
        [120, 251],
        [232, 300],
        [312, 300],
        [325, 300],
        [318, 295],
        [296, 260],
        [352, 259],
        [106, 210],
        [197, 206],
        [99, 200],
    ]
    assert (
        trim_polygon(polygon, TEST_IMAGE["width"], TEST_IMAGE["height"])
        == expected_polygon
    )


def test_trim_polygon():
    polygon = (
        (12, 56),
        (29, 60),
        (35, 61),
        (42, 59),
        (58, 57),
        (65, 61),
        (72, 57),
        (12, 56),
    )
    expected_polygon = [
        [12, 56],
        [29, 60],
        [35, 61],
        [42, 59],
        [58, 57],
        [65, 61],
        [72, 57],
        [12, 56],
    ]
    assert (
        trim_polygon(polygon, TEST_IMAGE["width"], TEST_IMAGE["height"])
        == expected_polygon
    )


@pytest.mark.parametrize(
    ("angle", "mirrored", "updated_bounds", "reverse"),
    [
        (
            0,
            False,
            {"x": 295, "y": 11, "width": 111, "height": 47},  # upper right
            True,
        ),
        (
            90,
            False,
            {"x": 510, "y": 295, "width": 47, "height": 111},  # lower right
            True,
        ),
        (
            180,
            False,
            {"x": 9, "y": 510, "width": 111, "height": 47},  # lower left
            True,
        ),
        (
            270,
            False,
            {"x": 11, "y": 9, "width": 47, "height": 111},  # upper left
            True,
        ),
        (
            0,
            True,
            {"x": 9, "y": 11, "width": 111, "height": 47},  # upper left
            True,
        ),
        (
            90,
            True,
            {"x": 510, "y": 9, "width": 47, "height": 111},  # upper right
            True,
        ),
        (
            180,
            True,
            {"x": 295, "y": 510, "width": 111, "height": 47},  # lower right
            True,
        ),
        (
            270,
            True,
            {"x": 11, "y": 295, "width": 47, "height": 111},  # lower left
            True,
        ),
        (
            0,
            False,
            {"x": 295, "y": 11, "width": 111, "height": 47},  # upper right
            False,
        ),
        (
            90,
            False,
            {"x": 11, "y": 162, "width": 47, "height": 111},  # upper left
            False,
        ),
        (
            180,
            False,
            {"x": 9, "y": 510, "width": 111, "height": 47},  # lower left
            False,
        ),
        (
            270,
            False,
            {"x": 357, "y": 295, "width": 47, "height": 111},  # lower right
            False,
        ),
        (
            0,
            True,
            {"x": 9, "y": 11, "width": 111, "height": 47},  # upper left
            False,
        ),
        (
            90,
            True,
            {"x": 357, "y": 162, "width": 47, "height": 111},  # lower left
            False,
        ),
        (
            180,
            True,
            {"x": 295, "y": 510, "width": 111, "height": 47},  # lower right
            False,
        ),
        (
            270,
            True,
            {"x": 11, "y": 295, "width": 47, "height": 111},  # upper right
            False,
        ),
    ],
)
def test_revert_orientation(angle, mirrored, updated_bounds, reverse, tmp_path):
    """Test cases, for both Elements and CachedElements:
    - no rotation or orientation
    - rotation with 3 different angles (90, 180, 270)
    - rotation + mirror with 4 angles (0, 90, 180, 270)
    """
    child_polygon = [[295, 11], [295, 58], [406, 58], [406, 11], [295, 11]]

    # Setup cache db to test with CachedElements
    db_path = f"{tmp_path}/db.sqlite"
    init_cache_db(db_path)
    create_tables()

    image_polygon = [
        [0, 0],
        [0, 568],
        [415, 568],
        [415, 0],
        [0, 0],
    ]
    element = Element(
        {
            "mirrored": mirrored,
            "rotation_angle": angle,
            "zone": {"polygon": image_polygon},
        }
    )
    cached_element = CachedElement.create(
        id=uuid.uuid4(),
        type="paragraph",
        polygon=image_polygon,
        mirrored=mirrored,
        rotation_angle=angle,
    )

    assert polygon_bounding_box(
        revert_orientation(element, child_polygon, reverse=reverse)
    ) == BoundingBox(**updated_bounds)

    assert polygon_bounding_box(
        revert_orientation(cached_element, child_polygon, reverse=reverse)
    ) == BoundingBox(**updated_bounds)


def test_download_image_retry_with_max(responses):
    base_url = "https://blabla.com/iiif/2/image_path.jpg/231,699,2789,3659/{size}/0/default.jpg"

    full_url = base_url.format(size=IIIF_FULL)
    max_url = base_url.format(size=IIIF_MAX)

    # Full size gives an error
    responses.add(
        responses.GET,
        full_url,
        status=400,
    )
    # Max size works
    responses.add(
        responses.GET,
        max_url,
        status=200,
        body=TILE.read_bytes(),
    )

    image = download_image(full_url)
    assert image

    # We try 3 times with the first URL
    # Then the first try with the new URL is successful
    assert len(responses.calls) == 4
    assert list(map(attrgetter("request.url"), responses.calls)) == [full_url] * 3 + [
        max_url
    ]


def test_upload_image_retries(responses):
    dest_url = "https://blabla.com/iiif/2/image_path.jpg/full/full/0/default.jpg"
    responses.add(
        responses.PUT,
        dest_url,
        status=400,
    )

    image = Image.open(FULL_IMAGE).convert("RGB")
    with pytest.raises(
        HTTPError, match=f"400 Client Error: Bad Request for url: {dest_url}"
    ):
        upload_image(image, dest_url)

    # We try 3 times
    assert len(responses.calls) == 3
    assert list(map(attrgetter("request.url"), responses.calls)) == [dest_url] * 3


def test_upload_image(responses):
    dest_url = "https://blabla.com/iiif/2/image_path.jpg/full/full/0/default.jpg"
    responses.add(
        responses.PUT,
        dest_url,
        status=200,
    )

    image = Image.open(FULL_IMAGE).convert("RGB")
    resp = upload_image(image, dest_url)
    assert resp

    assert len(responses.calls) == 1
    assert list(map(attrgetter("request.url"), responses.calls)) == [dest_url]


@pytest.mark.parametrize(
    (
        "max_pixels_short",
        "max_pixels_long",
        "max_bytes",
        "expected_sizes",
        "expected_logs",
    ),
    [
        # No limits provided
        (
            None,
            None,
            None,
            [(2000, 3000), (1000, 1500), (200, 300)],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (logging.WARNING, "The image was resized to (1000 x 1500)."),
                (logging.WARNING, "The image was resized to (200 x 300)."),
            ],
        ),
        # Image already under all three limits
        (
            10000,
            10000,
            4000000,  # 4MB
            [(2000, 3000), (1000, 1500), (200, 300)],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (logging.WARNING, "The image was resized to (1000 x 1500)."),
                (logging.WARNING, "The image was resized to (200 x 300)."),
            ],
        ),
        # Image above the "short side in pixels" limit
        (
            1000,
            None,
            None,
            [(1000, 1500), (500, 750), (100, 150)],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (
                    logging.WARNING,
                    "Maximum image dimensions supported are (1000 x 3000).",
                ),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (1000 x 1500)."),
                (logging.WARNING, "The image was resized to (500 x 750)."),
                (logging.WARNING, "The image was resized to (100 x 150)."),
            ],
        ),
        # Image above the "long side in pixels" limit
        (
            None,
            2000,
            None,
            [(1333, 2000), (667, 1000), (133, 200)],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (
                    logging.WARNING,
                    "Maximum image dimensions supported are (2000 x 2000).",
                ),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (1333 x 2000)."),
                (logging.WARNING, "The image was resized to (667 x 1000)."),
                (logging.WARNING, "The image was resized to (133 x 200)."),
            ],
        ),
        # Image above the "size in bytes" limit
        (
            None,
            None,
            100000,  # 100kB
            [(200, 300)],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (logging.WARNING, "The image size is 1.3 MB."),
                (logging.WARNING, "Maximum image input size supported is 100.0 kB."),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (1000 x 1500)."),
                (logging.WARNING, "The image size is 214.2 kB."),
                (logging.WARNING, "Maximum image input size supported is 100.0 kB."),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (200 x 300)."),
            ],
        ),
        # Image above all three limits
        (
            1000,
            2000,
            100000,  # 100kB
            [(500, 750), (100, 150)],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (
                    logging.WARNING,
                    "Maximum image dimensions supported are (1000 x 2000).",
                ),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (1000 x 1500)."),
                (logging.WARNING, "The image size is 214.2 kB."),
                (logging.WARNING, "Maximum image input size supported is 100.0 kB."),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (500 x 750)."),
                (logging.WARNING, "The image was resized to (100 x 150)."),
            ],
        ),
        # Image always above all three limits
        (
            50,
            50,
            50,  # 50B
            [],
            [
                (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
                (logging.WARNING, "Maximum image dimensions supported are (50 x 50)."),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (33 x 50)."),
                (logging.WARNING, "The image size is 1.0 kB."),
                (logging.WARNING, "Maximum image input size supported is 50 Bytes."),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (17 x 25)."),
                (logging.WARNING, "The image size is 785 Bytes."),
                (logging.WARNING, "Maximum image input size supported is 50 Bytes."),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (3 x 5)."),
                (logging.WARNING, "The image size is 689 Bytes."),
                (logging.WARNING, "Maximum image input size supported is 50 Bytes."),
                (logging.WARNING, "The image will be resized."),
            ],
        ),
    ],
)
def test_resized_images_portrait_format(
    monkeypatch,
    max_pixels_short,
    max_pixels_long,
    max_bytes,
    expected_sizes,
    expected_logs,
    mock_page,
    caplog,
):
    monkeypatch.setattr(arkindex_worker.image, "IMAGE_RATIOS", [1.0, 0.5, 0.1])

    # Short side is the width, long side is the height
    assert mock_page.zone.image.width < mock_page.zone.image.height

    assert [
        Image.open(image).size
        for image in resized_images(
            element=mock_page,
            max_pixels_short=max_pixels_short,
            max_pixels_long=max_pixels_long,
            max_bytes=max_bytes,
        )
    ] == expected_sizes

    assert [
        (record.levelno, record.message) for record in caplog.records
    ] == expected_logs


@pytest.mark.parametrize(
    ("max_pixels_short", "max_pixels_long", "expected_sizes", "expected_logs"),
    [
        # Image above the "short side in pixels" limit
        (
            1000,
            None,
            [(1500, 1000), (750, 500), (150, 100)],
            [
                (logging.INFO, "This element's image dimensions are (3000 x 2000)."),
                (
                    logging.WARNING,
                    "Maximum image dimensions supported are (3000 x 1000).",
                ),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (1500 x 1000)."),
                (logging.WARNING, "The image was resized to (750 x 500)."),
                (logging.WARNING, "The image was resized to (150 x 100)."),
            ],
        ),
        # Image above the "long side in pixels" limit
        (
            None,
            2000,
            [(2000, 1333), (1000, 667), (200, 133)],
            [
                (logging.INFO, "This element's image dimensions are (3000 x 2000)."),
                (
                    logging.WARNING,
                    "Maximum image dimensions supported are (2000 x 2000).",
                ),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (2000 x 1333)."),
                (logging.WARNING, "The image was resized to (1000 x 667)."),
                (logging.WARNING, "The image was resized to (200 x 133)."),
            ],
        ),
        # Image above the two pixels limits
        (
            1000,
            2000,
            [(1500, 1000), (750, 500), (150, 100)],
            [
                (logging.INFO, "This element's image dimensions are (3000 x 2000)."),
                (
                    logging.WARNING,
                    "Maximum image dimensions supported are (2000 x 1000).",
                ),
                (logging.WARNING, "The image will be resized."),
                (logging.WARNING, "The image was resized to (1500 x 1000)."),
                (logging.WARNING, "The image was resized to (750 x 500)."),
                (logging.WARNING, "The image was resized to (150 x 100)."),
            ],
        ),
    ],
)
def test_resized_images_landscape_format(
    monkeypatch,
    max_pixels_short,
    max_pixels_long,
    expected_sizes,
    expected_logs,
    mock_page,
    caplog,
):
    monkeypatch.setattr(arkindex_worker.image, "IMAGE_RATIOS", [1.0, 0.5, 0.1])

    # Short side is the height, long side is the width
    mock_page.zone = {
        "polygon": [[0, 0], [3000, 0], [3000, 2000], [0, 2000], [0, 0]],
        "image": {"width": 3000, "height": 2000},
    }
    assert mock_page.zone.image.height < mock_page.zone.image.width

    assert [
        Image.open(image).size
        for image in resized_images(
            element=mock_page,
            max_pixels_short=max_pixels_short,
            max_pixels_long=max_pixels_long,
            max_bytes=None,
        )
    ] == expected_sizes

    assert [
        (record.levelno, record.message) for record in caplog.records
    ] == expected_logs


def test_resized_images_use_base64(monkeypatch, mock_page, caplog):
    monkeypatch.setattr(arkindex_worker.image, "IMAGE_RATIOS", [1.0, 0.5, 0.25, 0.1])

    assert list(
        map(
            len,
            resized_images(
                element=mock_page,
                max_pixels_short=None,
                max_pixels_long=None,
                max_bytes=100000,
                use_base64=True,
            ),
        )
    ) == [65280, 11892]

    assert [(record.levelno, record.message) for record in caplog.records] == [
        (logging.INFO, "This element's image dimensions are (2000 x 3000)."),
        (logging.WARNING, "The image size is 1.7 MB."),
        (logging.WARNING, "Maximum image input size supported is 100.0 kB."),
        (logging.WARNING, "The image will be resized."),
        (logging.WARNING, "The image was resized to (1000 x 1500)."),
        (logging.WARNING, "The image size is 285.6 kB."),
        (logging.WARNING, "Maximum image input size supported is 100.0 kB."),
        (logging.WARNING, "The image will be resized."),
        (logging.WARNING, "The image was resized to (500 x 750)."),
        (logging.WARNING, "The image was resized to (200 x 300)."),
    ]



import hashlib
import json
import os
import sys
import time
from uuid import UUID

import pytest
import yaml
from peewee import SqliteDatabase

from arkindex.mock import MockApiClient
from arkindex_worker.cache import (
    MODELS,
    SQL_VERSION,
    CachedElement,
    CachedImage,
    CachedTranscription,
    Version,
    create_tables,
    create_version_table,
    init_cache_db,
)
from arkindex_worker.models import Artifact, Dataset, Set
from arkindex_worker.worker import (
    BaseWorker,
    DatasetWorker,
    ElementsWorker,
    ProcessMode,
)
from arkindex_worker.worker.dataset import DatasetState
from arkindex_worker.worker.transcription import TextOrientation
from tests import CORPUS_ID, SAMPLES_DIR

__yaml_cache = {}

ARKINDEX_API_SCHEMA_URL_VAR = "ARKINDEX_API_SCHEMA_URL"


@pytest.fixture(autouse=True)
def _disable_sleep(monkeypatch):
    """
    Do not sleep at all in between API executions
    when errors occur in unit tests.
    This speeds up the test execution a lot
    """
    monkeypatch.setattr(time, "sleep", lambda x: None)


@pytest.fixture
def _cache_yaml(monkeypatch):
    """
    Cache all calls to yaml.safe_load in order to speedup
    every test cases that load the OpenAPI schema
    """
    # Keep a reference towards the original function
    _original_yaml_load = yaml.safe_load

    def _cached_yaml_load(yaml_payload):
        # Create a unique cache key for direct YAML strings
        # and file descriptors
        if isinstance(yaml_payload, str):
            yaml_payload = yaml_payload.encode("utf-8")
        if isinstance(yaml_payload, bytes):
            key = hashlib.md5(yaml_payload).hexdigest()
        else:
            key = yaml_payload.name

        # Cache result
        if key not in __yaml_cache:
            __yaml_cache[key] = _original_yaml_load(yaml_payload)

        return __yaml_cache[key]

    monkeypatch.setattr(yaml, "safe_load", _cached_yaml_load)


@pytest.fixture(autouse=True)
def _setup_api(responses, monkeypatch, _cache_yaml):
    # Always use the environment variable first
    schema_url = os.environ.get(ARKINDEX_API_SCHEMA_URL_VAR)

    # Fallback to prod environment
    if schema_url is None:
        schema_url = "https://arkindex.teklia.com/api/v1/openapi/?format=json"
        monkeypatch.setenv(ARKINDEX_API_SCHEMA_URL_VAR, schema_url)

    # Allow accessing remote API schemas
    responses.add_passthru(schema_url)

    # Force api requests on a dummy server with dummy credentials
    monkeypatch.setenv("ARKINDEX_API_URL", "http://testserver/api/v1")
    monkeypatch.setenv("ARKINDEX_API_TOKEN", "unittest1234")


@pytest.fixture
def setup_mock_api_client(monkeypatch):
    def mock_setup_api_client(self):
        self.api_client = MockApiClient(
            schema_url=os.getenv(ARKINDEX_API_SCHEMA_URL_VAR)
        )

    monkeypatch.setattr(BaseWorker, "setup_api_client", mock_setup_api_client)


@pytest.fixture(autouse=True)
def _give_env_variable(monkeypatch):
    """Defines required environment variables"""
    monkeypatch.setenv("ARKINDEX_WORKER_RUN_ID", "56785678-5678-5678-5678-567856785678")


@pytest.fixture
def _mock_worker_run_api(responses):
    """Provide a mock API response to get worker run information"""
    payload = {
        "id": "56785678-5678-5678-5678-567856785678",
        "parents": [],
        "worker_version": {
            "id": "12341234-1234-1234-1234-123412341234",
            "configuration": {
                "docker": {"image": "python:3"},
                "configuration": {"someKey": "someValue"},
                "secrets": [],
            },
            "revision": {
                "hash": "deadbeef1234",
                "name": "some git revision",
            },
            "version": None,
            "docker_image": None,
            "docker_image_iid": "python:3",
            "docker_image_name": None,
            "state": "created",
            "gpu_usage": "disabled",
            "model_usage": "disabled",
            "worker": {
                "id": "deadbeef-1234-5678-1234-worker",
                "name": "Fake worker",
                "slug": "fake_worker",
                "type": "classifier",
            },
        },
        "configuration": {
            "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
            "name": "string",
            "configuration": {},
            "archived": False,
        },
        "model_version": None,
        "process": {
            "name": None,
            "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
            "state": "running",
            "mode": "workers",
            "corpus": CORPUS_ID,
            "use_cache": False,
            "activity_state": "ready",
            "model_id": None,
            "train_folder_id": None,
            "validation_folder_id": None,
            "test_folder_id": None,
            "skip_elements_json": False,
        },
        "summary": "Worker Fake worker @ 123412",
    }

    responses.add(
        responses.GET,
        "http://testserver/api/v1/process/workers/56785678-5678-5678-5678-567856785678/",
        status=200,
        body=json.dumps(payload),
        content_type="application/json",
    )

    # By default, stick to classic configuration
    responses.add(
        responses.GET,
        "http://testserver/api/v1/workers/runs/56785678-5678-5678-5678-567856785678/configuration/",
        status=400,
    )


@pytest.fixture
def _mock_worker_run_no_revision_api(responses):
    """Provide a mock API response to get worker run not linked to a revision information"""
    payload = {
        "id": "56785678-5678-5678-5678-567856785678",
        "parents": [],
        "worker_version": {
            "id": "12341234-1234-1234-1234-123412341234",
            "configuration": {
                "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
                "docker": {"image": "python:3"},
                "configuration": {"someKey": "someValue"},
                "secrets": [],
            },
            "revision": None,
            "version": 1,
            "docker_image": None,
            "docker_image_iid": "python:3",
            "docker_image_name": None,
            "state": "created",
            "gpu_usage": "disabled",
            "model_usage": "disabled",
            "worker": {
                "id": "deadbeef-1234-5678-1234-worker",
                "name": "Fake worker",
                "slug": "fake_worker",
                "type": "classifier",
            },
        },
        "configuration": {
            "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
            "name": "string",
            "configuration": {},
            "archived": False,
        },
        "model_version": None,
        "process": {
            "name": None,
            "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
            "state": "running",
            "mode": "workers",
            "corpus": CORPUS_ID,
            "use_cache": False,
            "activity_state": "ready",
            "model_id": None,
            "train_folder_id": None,
            "validation_folder_id": None,
            "test_folder_id": None,
        },
        "summary": "Worker Fake worker @ 1",
    }

    responses.add(
        responses.GET,
        "http://testserver/api/v1/process/workers/56785678-5678-5678-5678-567856785678/",
        status=200,
        body=json.dumps(payload),
        content_type="application/json",
    )


@pytest.fixture
def mock_base_worker_modern_conf(mocker, responses):
    """
    Provide a base worker to test modern configuration with (not provided in the fixture)
    """
    worker = BaseWorker()
    mocker.patch.object(sys, "argv")
    worker.args = worker.parser.parse_args()

    payload = {
        "id": "56785678-5678-5678-5678-567856785678",
        "parents": [],
        "worker_version": {
            "id": "12341234-1234-1234-1234-123412341234",
            "worker": {
                "id": "deadbeef-1234-5678-1234-worker",
                "name": "Fake worker",
                "slug": "fake_worker",
                "type": "classifier",
            },
            "revision": {"hash": "deadbeef1234"},
            "configuration": {
                "configuration": {"extra_key1": "not showing up"},
                "user_configuration": {"extra_key2": "not showing up"},
            },
        },
        "configuration": {
            "id": "af0daaf4-983e-4703-a7ed-a10f146d6684",
            "name": "my-userconfig",
            "configuration": {
                "extra_key3": "not showing up",
            },
        },
        "model_version": None,
        "process": {
            "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
            "corpus": CORPUS_ID,
        },
        "summary": "Worker Fake worker @ 123412",
    }
    responses.add(
        responses.GET,
        "http://testserver/api/v1/process/workers/56785678-5678-5678-5678-567856785678/",
        status=200,
        json=payload,
    )

    return worker


@pytest.fixture
def _mock_activity_calls(responses):
    """
    Mock responses when updating the activity state for multiple element of the same version
    """
    responses.add(
        responses.PUT,
        "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        status=200,
    )


@pytest.fixture
def mock_elements_worker(monkeypatch, _mock_worker_run_api):
    """Build and configure an ElementsWorker with fixed CLI parameters to avoid issues with pytest"""
    monkeypatch.setattr(sys, "argv", ["worker"])
    worker = ElementsWorker()
    worker.configure()
    return worker


@pytest.fixture
def mock_elements_worker_read_only(monkeypatch):
    """Build and configure an ElementsWorker with fixed CLI parameters to avoid issues with pytest"""
    monkeypatch.setattr(sys, "argv", ["worker", "--dev"])
    worker = ElementsWorker()
    worker.configure()
    return worker


@pytest.fixture
def mock_elements_worker_with_list(monkeypatch, responses, mock_elements_worker):
    """
    Mock a worker instance to list and retrieve a single element
    """
    monkeypatch.setattr(mock_elements_worker, "get_elements", lambda: ["1234-deadbeef"])
    responses.add(
        responses.GET,
        "http://testserver/api/v1/element/1234-deadbeef/",
        status=200,
        json={
            "id": "1234-deadbeef",
            "type": "page",
            "name": "Test Page n°1",
        },
    )
    return mock_elements_worker


@pytest.fixture
def mock_elements_worker_consume_wa(monkeypatch, responses, mock_elements_worker):
    """
    Mock a worker instance to use StartWorkerActivity to consume worker activities
    instead of reading a JSON file
    """

    # Enable consume worker activities through the process configuration
    responses.replace(
        responses.GET,
        "http://testserver/api/v1/process/workers/56785678-5678-5678-5678-567856785678/",
        status=200,
        json={
            "id": "56785678-5678-5678-5678-567856785678",
            "parents": [],
            "worker_version": {
                "id": "12341234-1234-1234-1234-123412341234",
                "configuration": {
                    "docker": {"image": "python:3"},
                    "configuration": {"someKey": "someValue"},
                    "secrets": [],
                },
                "worker": {
                    "id": "deadbeef-1234-5678-1234-worker",
                    "name": "Fake worker",
                    "slug": "fake_worker",
                    "type": "classifier",
                },
            },
            "configuration": None,
            "model_version": None,
            "process": {
                "name": None,
                "id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
                "state": "running",
                "mode": "workers",
                "corpus": CORPUS_ID,
                "use_cache": False,
                "activity_state": "ready",
                "model_id": None,
                "train_folder_id": None,
                "validation_folder_id": None,
                "test_folder_id": None,
                "skip_elements_json": True,
            },
            "summary": "Worker Fake worker @ 123412",
        },
    )

    # Call configure again to use updated process infos
    mock_elements_worker.configure()

    return mock_elements_worker


@pytest.fixture
def mock_cache_db(tmp_path):
    cache_path = tmp_path / "db.sqlite"

    init_cache_db(cache_path)
    create_version_table()
    create_tables()

    return cache_path


@pytest.fixture
def mock_base_worker_with_cache(monkeypatch, _mock_worker_run_api):
    """Build a BaseWorker using SQLite cache, also mocking a PONOS_TASK"""
    monkeypatch.setattr(sys, "argv", ["worker"])

    monkeypatch.setenv("PONOS_TASK", "my_task")
    return BaseWorker(support_cache=True)


@pytest.fixture
def mock_elements_worker_with_cache(monkeypatch, mock_cache_db, _mock_worker_run_api):
    """Build and configure an ElementsWorker using SQLite cache with fixed CLI parameters to avoid issues with pytest"""
    monkeypatch.setattr(sys, "argv", ["worker", "-d", str(mock_cache_db)])

    worker = ElementsWorker(support_cache=True)
    worker.configure()
    return worker


@pytest.fixture
def model_file_dir():
    return SAMPLES_DIR / "model_files"


@pytest.fixture
def model_file_dir_with_subfolder():
    return SAMPLES_DIR / "root_folder"


@pytest.fixture
def fake_dummy_worker(setup_mock_api_client):
    return ElementsWorker()


@pytest.fixture
def _mock_cached_elements(mock_cache_db):
    """Insert few elements in local cache"""
    CachedElement.create(
        id=UUID("99999999-9999-9999-9999-999999999999"),
        parent_id=None,
        type="something",
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=None,
        worker_run_id=None,
    )
    CachedElement.create(
        id=UUID("12341234-1234-1234-1234-123412341234"),
        parent_id=UUID("99999999-9999-9999-9999-999999999999"),
        type="double_page",
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("11111111-1111-1111-1111-111111111111"),
        parent_id=UUID("12341234-1234-1234-1234-123412341234"),
        type="something",
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("22222222-2222-2222-2222-222222222222"),
        parent_id=UUID("12341234-1234-1234-1234-123412341234"),
        type="page",
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("33333333-3333-3333-3333-333333333333"),
        parent_id=UUID("12341234-1234-1234-1234-123412341234"),
        type="paragraph",
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=None,
        worker_run_id=None,
    )
    assert CachedElement.select().count() == 5


@pytest.fixture
def _mock_cached_images(mock_cache_db):
    """Insert few elements in local cache"""
    CachedImage.create(
        id=UUID("99999999-9999-9999-9999-999999999999"),
        width=1250,
        height=2500,
        url="http://testserver/iiif/3/image",
    )
    assert CachedImage.select().count() == 1


@pytest.fixture
def _mock_cached_transcriptions(mock_cache_db):
    """Insert few transcriptions in local cache, on a shared element"""
    CachedElement.create(
        id=UUID("11111111-1111-1111-1111-111111111111"),
        type="page",
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("22222222-2222-2222-2222-222222222222"),
        type="something_else",
        parent_id=UUID("11111111-1111-1111-1111-111111111111"),
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("33333333-3333-3333-3333-333333333333"),
        type="page",
        parent_id=UUID("11111111-1111-1111-1111-111111111111"),
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("44444444-4444-4444-4444-444444444444"),
        type="something_else",
        parent_id=UUID("22222222-2222-2222-2222-222222222222"),
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedElement.create(
        id=UUID("55555555-5555-5555-5555-555555555555"),
        type="something_else",
        parent_id=UUID("44444444-4444-4444-4444-444444444444"),
        polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedTranscription.create(
        id=UUID("11111111-1111-1111-1111-111111111111"),
        element_id=UUID("11111111-1111-1111-1111-111111111111"),
        text="This",
        confidence=0.42,
        orientation=TextOrientation.HorizontalLeftToRight,
        worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
    )
    CachedTranscription.create(
        id=UUID("22222222-2222-2222-2222-222222222222"),
        element_id=UUID("22222222-2222-2222-2222-222222222222"),
        text="is",
        confidence=0.42,
        orientation=TextOrientation.HorizontalLeftToRight,
        worker_version_id=UUID("90129012-9012-9012-9012-901290129012"),
    )
    CachedTranscription.create(
        id=UUID("33333333-3333-3333-3333-333333333333"),
        element_id=UUID("33333333-3333-3333-3333-333333333333"),
        text="a",
        confidence=0.42,
        orientation=TextOrientation.HorizontalLeftToRight,
        worker_version_id=UUID("90129012-9012-9012-9012-901290129012"),
    )
    CachedTranscription.create(
        id=UUID("44444444-4444-4444-4444-444444444444"),
        element_id=UUID("44444444-4444-4444-4444-444444444444"),
        text="good",
        confidence=0.42,
        orientation=TextOrientation.HorizontalLeftToRight,
        worker_version_id=UUID("90129012-9012-9012-9012-901290129012"),
    )
    CachedTranscription.create(
        id=UUID("55555555-5555-5555-5555-555555555555"),
        element_id=UUID("55555555-5555-5555-5555-555555555555"),
        text="test",
        confidence=0.42,
        orientation=TextOrientation.HorizontalLeftToRight,
        worker_version_id=UUID("90129012-9012-9012-9012-901290129012"),
    )
    CachedTranscription.create(
        id=UUID("66666666-6666-6666-6666-666666666666"),
        element_id=UUID("11111111-1111-1111-1111-111111111111"),
        text="This is a manual one",
        confidence=0.42,
        orientation=TextOrientation.HorizontalLeftToRight,
        worker_version_id=None,
        worker_run_id=None,
    )


@pytest.fixture
def mock_databases(tmp_path):
    """
    Initialize several temporary databases
    to help testing the merge algorithm
    """
    out = {}
    for name in ("target", "first", "second", "conflict", "chunk_42"):
        # Build a local database in sub directory
        # for each name required
        filename = "db_42.sqlite" if name == "chunk_42" else "db.sqlite"
        path = tmp_path / name / filename
        (tmp_path / name).mkdir()
        local_db = SqliteDatabase(path)
        with local_db.bind_ctx(MODELS + [Version]):
            # Create tables on the current local database
            # by binding temporarily the models on that database
            local_db.create_tables([Version])
            Version.create(version=SQL_VERSION)
            local_db.create_tables(MODELS)
        out[name] = {"path": path, "db": local_db}

    # Add an element in first parent database
    with out["first"]["db"].bind_ctx(MODELS):
        CachedElement.create(
            id=UUID("12341234-1234-1234-1234-123412341234"),
            type="page",
            polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
            worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        )
        CachedElement.create(
            id=UUID("56785678-5678-5678-5678-567856785678"),
            type="page",
            polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
            worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        )

    # Add another element with a transcription in second parent database
    with out["second"]["db"].bind_ctx(MODELS):
        CachedElement.create(
            id=UUID("42424242-4242-4242-4242-424242424242"),
            type="page",
            polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
            worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        )
        CachedTranscription.create(
            id=UUID("11111111-1111-1111-1111-111111111111"),
            element_id=UUID("42424242-4242-4242-4242-424242424242"),
            text="Hello!",
            confidence=0.42,
            orientation=TextOrientation.HorizontalLeftToRight,
            worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        )

    # Add a conflicting element
    with out["conflict"]["db"].bind_ctx(MODELS):
        CachedElement.create(
            id=UUID("42424242-4242-4242-4242-424242424242"),
            type="page",
            polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
            initial=True,
        )
        CachedTranscription.create(
            id=UUID("22222222-2222-2222-2222-222222222222"),
            element_id=UUID("42424242-4242-4242-4242-424242424242"),
            text="Hello again neighbor !",
            confidence=0.42,
            orientation=TextOrientation.HorizontalLeftToRight,
            worker_version_id=UUID("56785678-5678-5678-5678-567856785678"),
        )

    # Add an element in chunk parent database
    with out["chunk_42"]["db"].bind_ctx(MODELS):
        CachedElement.create(
            id=UUID("42424242-4242-4242-4242-424242424242"),
            type="page",
            polygon="[[1, 1], [2, 2], [2, 1], [1, 2]]",
            initial=True,
        )

    return out


@pytest.fixture
def default_dataset():
    return Dataset(
        {
            "id": "dataset_id",
            "name": "My dataset",
            "description": "A super dataset built by me",
            "sets": ["set_1", "set_2", "set_3", "set_4"],
            "state": DatasetState.Open.value,
            "corpus_id": "corpus_id",
            "creator": "creator@teklia.com",
            "task_id": "11111111-1111-1111-1111-111111111111",
            "created": "2000-01-01T00:00:00Z",
            "updated": "2000-01-01T00:00:00Z",
        }
    )


@pytest.fixture
def default_train_set(default_dataset):
    return Set(name="train", dataset=default_dataset)


@pytest.fixture
def mock_dataset_worker(monkeypatch, mocker, _mock_worker_run_api):
    monkeypatch.setenv("PONOS_TASK", "my_task")
    mocker.patch.object(sys, "argv", ["worker"])

    dataset_worker = DatasetWorker()
    dataset_worker.configure()

    # Update process mode
    dataset_worker.process_information["mode"] = ProcessMode.Dataset

    assert not dataset_worker.is_read_only

    return dataset_worker


@pytest.fixture
def mock_dev_dataset_worker(mocker):
    mocker.patch.object(
        sys,
        "argv",
        [
            "worker",
            "--dev",
            "--set",
            "11111111-1111-1111-1111-111111111111:train",
            "11111111-1111-1111-1111-111111111111:val",
            "22222222-2222-2222-2222-222222222222:my_set",
        ],
    )

    dataset_worker = DatasetWorker()
    dataset_worker.configure()

    assert dataset_worker.args.dev is True
    assert dataset_worker.process_information is None
    assert dataset_worker.is_read_only is True

    return dataset_worker


@pytest.fixture
def default_artifact():
    return Artifact(
        **{
            "id": "artifact_id",
            "path": "dataset_id.tar.zst",
            "size": 42,
            "content_type": "application/zstd",
            "s3_put_url": None,
            "created": "2000-01-01T00:00:00Z",
            "updated": "2000-01-01T00:00:00Z",
        }
    )



"""
BaseWorker methods for training.
"""

import functools
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import NewType
from uuid import UUID

from arkindex_worker import logger
from arkindex_worker.utils import close_delete_file, create_tar_zst_archive
from teklia_toolbox.uploads import MultipartUpload

DirPath = NewType("DirPath", Path)
"""Path to a directory"""

Hash = NewType("Hash", str)
"""MD5 Hash"""

FileSize = NewType("FileSize", int)
"""File size"""


@contextmanager
def create_archive(path: DirPath) -> Generator[tuple[Path, FileSize, Hash]]:
    """
    Create a tar archive from the files at the given location then compress it to a zst archive.

    Yield its location, its size and its hash.

    :param path: Create a compressed tar archive from the files
    :returns: The location of the created archive, its size and its hash
    """
    assert path.is_dir(), "create_archive needs a directory"

    zst_descriptor, zst_archive, archive_hash = create_tar_zst_archive(path)

    # Get content hash, archive size and hash
    yield zst_archive, zst_archive.stat().st_size, archive_hash

    # Remove the zst archive
    close_delete_file(zst_descriptor, zst_archive)


def build_clean_payload(**kwargs):
    """
    Remove null attributes from an API body payload
    """
    return {key: value for key, value in kwargs.items() if value is not None}


def skip_if_read_only(func):
    """
    Return shortly in case the is_read_only property is evaluated to True
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if getattr(self, "is_read_only", False):
            logger.warning(
                "Cannot perform this operation as the worker is in read-only mode"
            )
            return
        return func(self, *args, **kwargs)

    return wrapper


class TrainingMixin:
    """
    A mixin helper to create a new model version easily.
    You may use `publish_model_version` to publish a ready model version directly, or
    separately create the model version then publish it (e.g to store training metrics).
    Stores the currently handled model version as `self.model_version`.
    """

    model_version = None

    @property
    def is_finetuning(self) -> bool:
        """
        Whether or not this worker is fine-tuning an existing model version.
        """
        return bool(self.model_version_id)

    @skip_if_read_only
    def publish_model_version(
        self,
        model_path: DirPath,
        model_id: str,
        tag: str | None = None,
        description: str | None = None,
        configuration: dict | None = None,
        parent: str | UUID | None = None,
    ):
        """
        Publish a unique version of a model in Arkindex, identified by its hash.
        In case the `create_model_version` method has been called, reuses that model
        instead of creating a new one.

        :param model_path: Path to the directory containing the model version's files.
        :param model_id: ID of the model
        :param tag: Tag of the model version
        :param description: Description of the model version
        :param configuration: Configuration of the model version
        :param parent: ID of the parent model version
        """

        configuration = configuration or {}

        # Create the zst archive, get its hash and size
        with create_archive(path=model_path) as (
            path_to_archive,
            size,
            hash,
        ):
            # Update an existing model version with hash, size and any other defined attribute
            if self.model_version:
                assert self.model_version.get("model_id") == model_id, (
                    "Given `model_id` does not match the current model version"
                )
                self.update_model_version(
                    size=size,
                    archive_hash=hash,
                    tag=tag,
                    description=description,
                    configuration=configuration,
                    parent=parent,
                )

            # Create a new model version with hash and size
            else:
                self.create_model_version(
                    model_id=model_id,
                    size=size,
                    archive_hash=hash,
                    tag=tag,
                    description=description,
                    configuration=configuration,
                    parent=parent,
                )

            # Upload the archive in multiple parts (supports huge files)
            self.upload_to_s3(path_to_archive)

    @skip_if_read_only
    def create_model_version(
        self,
        model_id: str,
        size: FileSize,
        archive_hash: Hash,
        tag: str | None = None,
        description: str | None = None,
        configuration: dict | None = None,
        parent: str | UUID | None = None,
    ):
        """
        Create a new version of the specified model with its base attributes.
        Once successfully created, the model version is accessible via `self.model_version`.

        :param size: Size of uploaded archive
        :param hash: MD5 hash of the uploaded archive
        :param tag: Tag of the model version
        :param description: Description of the model version
        :param configuration: Configuration of the model version
        :param parent: ID of the parent model version
        """
        assert not self.model_version, "A model version has already been created."

        configuration = configuration or {}
        self.model_version = self.api_client.request(
            "CreateModelVersion",
            id=model_id,
            body=build_clean_payload(
                size=size,
                archive_hash=archive_hash,
                tag=tag,
                description=description,
                configuration=configuration,
                parent=parent,
            ),
        )

        logger.info(
            f"Model version ({self.model_version['id']}) was successfully created."
        )

    @skip_if_read_only
    def update_model_version(
        self,
        size: FileSize,
        archive_hash: Hash,
        tag: str | None = None,
        description: str | None = None,
        configuration: dict | None = None,
        parent: str | UUID | None = None,
    ):
        """
        Update the current model version with the given attributes.

        :param size: Size of uploaded archive
        :param hash: MD5 hash of the uploaded archive
        :param tag: Tag of the model version
        :param description: Description of the model version
        :param configuration: Configuration of the model version
        :param parent: ID of the parent model version
        """
        assert self.model_version, "No model version has been created yet."
        self.model_version = self.api_client.request(
            "UpdateModelVersion",
            id=self.model_version["id"],
            body=build_clean_payload(
                size=size,
                archive_hash=archive_hash,
                tag=tag,
                description=description,
                configuration=configuration,
                parent=parent,
            ),
        )
        logger.info(
            f"Model version ({self.model_version['id']}) was successfully updated."
        )

    @skip_if_read_only
    def upload_to_s3(self, archive_path: Path) -> None:
        """
        Upload the archive of the model's files to an Amazon s3 compatible storage in multiple parts
        """
        assert self.model_version, (
            "You must create the model version before uploading an archive."
        )
        assert self.model_version["state"] != "Available", (
            "The model version is already marked as available."
        )

        multipart = MultipartUpload(
            client=self.api_client,
            file_path=archive_path,
            object_type="model_version",
            object_id=str(self.model_version["id"]),
        )
        try:
            multipart.upload()
            multipart.complete()
        except Exception:
            multipart.abort()
            raise
        else:
            logger.info(
                f"Model version ({self.model_version['id']}) archive was successfully uploaded and is now available."
            )



"""
Wrappers around API results to provide more convenient attribute access and IIIF helpers.
"""

import tempfile
from collections.abc import Generator
from contextlib import contextmanager

from PIL import Image
from requests import HTTPError

IMAGE_EXTENSION = "jpg"


class MagicDict(dict):
    """
    A dict whose items can be accessed like attributes.
    """

    def _magify(self, item):
        """
        Automagically convert lists and dicts to MagicDicts and lists of MagicDicts
        Allows for nested access: foo.bar.baz
        """
        if isinstance(item, Dataset):
            return item
        if isinstance(item, list):
            return list(map(self._magify, item))
        if isinstance(item, dict):
            return MagicDict(item)
        return item

    def __getitem__(self, item):
        item = super().__getitem__(item)
        return self._magify(item)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{name}'"
            ) from e

    def __setattr__(self, name, value):
        return super().__setitem__(name, value)

    def __delattr__(self, name):
        try:
            return super().__delattr__(name)
        except AttributeError:
            try:
                return super().__delitem__(name)
            except KeyError:
                pass
            raise

    def __dir__(self):
        return super().__dir__() + list(self.keys())


class Element(MagicDict):
    """
    Describes an Arkindex element.
    """

    def resize_zone_url(
        self, size: str = "full", extension: str = IMAGE_EXTENSION
    ) -> str:
        """
        Compute the URL of the image corresponding to the size
        :param size: Requested size
        :param extension: IIIF extension to download the image
        :return: The URL corresponding to the size
        """
        # Removing the `jpg` default extension at the end of the URL to use the provided one
        url = self.zone.url[:-3] + extension

        if size == "full":
            return url
        else:
            parts = url.split("/")
            parts[-3] = size
            return "/".join(parts)

    def image_url(
        self, size: str = "full", extension: str = IMAGE_EXTENSION
    ) -> str | None:
        """
        Build a URL to access the image.
        When possible, will return the S3 URL for images, so an ML worker can bypass IIIF servers.
        :param size: Subresolution of the image, following the syntax of the IIIF resize parameter.
        :param extension: IIIF extension to download the image.
        :returns: A URL to the image, or None if the element does not have an image.
        """
        if not self.get("zone"):
            return

        url = self.zone.image.get("s3_url")
        if url:
            return url

        # Use `max` instead of `full` for IIIF 3, since `full` was deprecated in 2.1 then removed in 3.0
        if self.zone.image.server.get("version", 2) == 3 and size == "full":
            size = "max"

        url = self.zone.image.url
        if not url.endswith("/"):
            url += "/"
        return f"{url}full/{size}/0/default.{extension}"

    @property
    def polygon(self) -> list[float]:
        """
        Access an Element's polygon.
        This is a shortcut to an Element's polygon, normally accessed via
        its zone field via `zone.polygon`. This is mostly done
        to facilitate access to this important field by matching
        the [CachedElement][arkindex_worker.cache.CachedElement].polygon field.
        """
        if not self.get("zone"):
            raise ValueError(f"Element {self.id} has no zone")
        return self.zone.polygon

    @property
    def requires_tiles(self) -> bool:
        """
        Whether or not downloading and combining IIIF tiles will be necessary
           to retrieve this element's image. Will be False if the element has no image.
        """
        from arkindex_worker.image import polygon_bounding_box

        if not self.get("zone") or self.zone.image.get("s3_url"):
            return False
        max_width = self.zone.image.server.max_width or float("inf")
        max_height = self.zone.image.server.max_height or float("inf")
        bounding_box = polygon_bounding_box(self.zone.polygon)
        return bounding_box.width > max_width or bounding_box.height > max_height

    def open_image(
        self,
        *args,
        max_width: int | None = None,
        max_height: int | None = None,
        use_full_image: bool | None = False,
        extension: str | None = IMAGE_EXTENSION,
        **kwargs,
    ) -> Image.Image:
        """
        Open this element's image using Pillow, rotating and mirroring it according
        to the ``rotation_angle`` and ``mirrored`` attributes.

        When tiling is not required to download the image, and no S3 URL is available
        to bypass IIIF servers, the image will be cropped to the rectangle bounding box
        of the ``zone.polygon`` attribute.

        Warns:
        ----
           This method implicitly applies the element's orientation to the image.

           If your process uses the returned image to find more polygons and send them
           back to Arkindex, use the [arkindex_worker.image.revert_orientation][]
           helper to undo the orientation on all polygons before sending them, as the
           Arkindex API expects unoriented polygons.

           Although not recommended, you can bypass this behavior by passing
           ``rotation_angle=0, mirrored=False`` as keyword arguments.


        Warns:
        ----
           If both, ``max_width`` and ``max_height`` are set, the image ratio is not preserved.


        :param max_width: The maximum width of the image.
        :param max_height: The maximum height of the image.
        :param use_full_image: Ignore the ``zone.polygon`` and always
           retrieve the image without cropping.
        :param extension: The extension to download the image.
        :param *args: Positional arguments passed to [arkindex_worker.image.open_image][].
        :param **kwargs: Keyword arguments passed to [arkindex_worker.image.open_image][].
        :raises ValueError: When the element does not have an image.
        :raises NotImplementedError: When the ``max_size`` parameter is set,
           but the IIIF server's configuration requires downloading and combining tiles
           to retrieve the image.
        :raises NotImplementedError: When an S3 URL has been used to download the image,
           but the URL has expired. Re-fetching the URL automatically is not supported.
        :return: A Pillow image.
        """
        from arkindex_worker.image import (
            download_tiles,
            open_image,
        )

        if not self.get("zone"):
            raise ValueError(f"Element {self.id} has no zone")

        if self.requires_tiles:
            if max_width is None and max_height is None:
                return download_tiles(self.zone.image.url)
            else:
                raise NotImplementedError

        if max_width is None and max_height is None:
            resize = "full"
        else:
            original_size = {"w": self.zone.image.width, "h": self.zone.image.height}
            # No resizing if the image is smaller than the wanted size.
            if (max_width is None or original_size["w"] <= max_width) and (
                max_height is None or original_size["h"] <= max_height
            ):
                resize = "full"
            # Resizing if the image is bigger than the wanted size.
            else:
                resize = f"{max_width or ''},{max_height or ''}"

        url = (
            self.image_url(resize, extension)
            if use_full_image
            else self.resize_zone_url(resize, extension)
        )

        try:
            return open_image(
                url,
                *args,
                rotation_angle=self.rotation_angle,
                mirrored=self.mirrored,
                **kwargs,
            )
        except HTTPError as e:
            if (
                self.zone.image.get("s3_url") is not None
                and e.response.status_code == 403
            ):
                # This element uses an S3 URL: the URL may have expired.
                # Call the API to get a fresh URL and try again
                # TODO: this should be done by the worker
                raise NotImplementedError from e
                return open_image(self.image_url(resize), *args, **kwargs)
            raise

    @contextmanager
    def open_image_tempfile(
        self, extension: str | None = IMAGE_EXTENSION, *args, **kwargs
    ) -> Generator[tempfile.NamedTemporaryFile, None, None]:
        """
        Get the element's image as a temporary file stored on the disk.
        To be used as a context manager.

        Example
        ----
        ```
        with element.open_image_tempfile() as f:
            ...
        ```

        :param format: File format to use the store the image on the disk.
           Must be a format supported by Pillow.
        :param *args: Positional arguments passed to [arkindex_worker.image.open_image][].
        :param **kwargs: Keyword arguments passed to [arkindex_worker.image.open_image][].

        """
        PIL_format = "jpeg" if extension == IMAGE_EXTENSION else extension

        with tempfile.NamedTemporaryFile(suffix=f".{extension}") as f:
            self.open_image(*args, extension=extension, **kwargs).save(
                f, format=PIL_format
            )
            yield f

    def __str__(self):
        if isinstance(self.type, dict):
            type_name = self.type["display_name"]
        else:
            type_name = str(self.type)
        return f"{type_name} {self.name} ({self.id})"


class ArkindexModel(MagicDict):
    def __str__(self):
        return f"{self.__class__.__name__} ({self.id})"


class Transcription(ArkindexModel):
    """
    Describes an Arkindex element's transcription.
    """


class Image(ArkindexModel):
    """
    Describes an Arkindex image.
    """


class Dataset(ArkindexModel):
    """
    Describes an Arkindex dataset.
    """

    @property
    def filepath(self) -> str:
        """
        Generic filepath to the Dataset compressed archive.
        """
        return f"{self.id}.tar.zst"


class Set(MagicDict):
    """
    Describes an Arkindex dataset set.
    """

    def __str__(self):
        # Not using ArkindexModel.__str__ as we do not retrieve the Set ID
        return f"{self.__class__.__name__} ({self.name}) from {self.dataset}"


class Artifact(ArkindexModel):
    """
    Describes an Arkindex artifact.
    """



"""
ElementsWorker methods for entities.
"""

from operator import itemgetter
from typing import TypedDict
from warnings import warn

from peewee import IntegrityError

from arkindex.exceptions import ErrorResponse
from arkindex_worker import logger
from arkindex_worker.cache import (
    CachedTranscriptionEntity,
    unsupported_cache,
)
from arkindex_worker.models import Transcription
from arkindex_worker.utils import pluralize


class Entity(TypedDict):
    type_id: str
    length: int
    offset: int
    confidence: float | None


class MissingEntityType(Exception):
    """
    Raised when the specified entity type was not found in the corpus and
    the worker cannot create it.
    """


class EntityMixin:
    def list_corpus_entity_types(self):
        """
        Loads available entity types in corpus.
        """
        self.entity_types = {
            entity_type["name"]: entity_type["id"]
            for entity_type in self.api_client.paginate(
                "ListCorpusEntityTypes", id=self.corpus_id
            )
        }
        count = len(self.entity_types)
        logger.info(
            f"Loaded {count} entity {pluralize('type', count)} in corpus ({self.corpus_id})."
        )

    @unsupported_cache
    def create_entity_type(self, name: str) -> None:
        """
        Create an entity type on the given corpus.

        :param name: Name of the entity type.
        """
        assert name and isinstance(name, str), (
            "name shouldn't be null and should be of type str"
        )

        try:
            entity_type = self.api_client.request(
                "CreateEntityType",
                body={
                    "name": name,
                    "corpus": self.corpus_id,
                },
            )
            self.entity_types[name] = entity_type["id"]
            logger.info(f"Created a new entity type with name `{name}`.")
        except ErrorResponse as e:
            # Only reload for 400 errors
            if e.status_code != 400:
                raise

            # Reload and make sure we have the element type now
            logger.warning(
                f"Unable to create the entity type `{name}`. Refreshing corpus entity types cache."
            )
            self.list_corpus_entity_types()
            assert name in self.entity_types, (
                f"Missing entity type `{name}` even after refreshing."
            )

    def check_required_entity_types(
        self, entity_types: list[str], create_missing: bool = True
    ) -> None:
        """
        Check that every entity type needed is available in the corpus.
        Missing ones may be created automatically if needed.

        :param entity_types: Entity type names to search.
        :param create_missing: Whether the missing types should be created. Defaults to True.
        :raises MissingEntityType: When an entity type is missing and cannot be created.
        """
        assert entity_types and isinstance(entity_types, list), (
            "entity_types shouldn't be null and should be of type list"
        )

        for index, entity_type in enumerate(entity_types):
            assert isinstance(entity_type, str), (
                f"Entity type at index {index} in entity_types: Should be of type str"
            )

        assert create_missing is not None and isinstance(create_missing, bool), (
            "create_missing shouldn't be null and should be of type bool"
        )

        if not self.entity_types:
            self.list_corpus_entity_types()

        for entity_type in entity_types:
            # Do nothing if the type already exists
            if entity_type in self.entity_types:
                continue

            # Do not create missing if not requested
            if not create_missing:
                raise MissingEntityType(
                    f"Entity type `{entity_type}` was not in the corpus."
                )

            # Create the type if non-existent
            self.create_entity_type(entity_type)

    def create_transcription_entity(
        self,
        transcription: Transcription,
        type_id: str,
        offset: int,
        length: int,
        confidence: float | None = None,
    ) -> dict[str, str | int] | None:
        """
        Create an entity on an existing transcription.
        If cache support is enabled, a `CachedTranscriptionEntity` will also be created.

        :param transcription: Transcription to create the entity on.
        :param type_id: UUID of the entity type.
        :param offset: Starting position of the entity in the transcription's text,
           as a 0-based index.
        :param length: Length of the entity in the transcription's text.
        :param confidence: Optional confidence score between 0 or 1.
        :returns: A dict as returned by the ``CreateTranscriptionEntity`` API endpoint,
           or None if the worker is in read-only mode.
        """
        assert transcription and isinstance(transcription, Transcription), (
            "transcription shouldn't be null and should be a Transcription"
        )
        assert type_id and isinstance(type_id, str), (
            "type_id shouldn't be null and should be of type str"
        )
        assert offset is not None and isinstance(offset, int) and offset >= 0, (
            "offset shouldn't be null and should be a positive integer"
        )
        assert length is not None and isinstance(length, int) and length > 0, (
            "length shouldn't be null and should be a strictly positive integer"
        )
        assert (
            confidence is None or isinstance(confidence, float) and 0 <= confidence <= 1
        ), "confidence should be null or a float in [0..1] range"
        if self.is_read_only:
            logger.warning(
                "Cannot create transcription entity as this worker is in read-only mode"
            )
            return

        body = {
            "type_id": type_id,
            "length": length,
            "offset": offset,
            "worker_run_id": self.worker_run_id,
        }
        if confidence is not None:
            body["confidence"] = confidence

        tr_entity = self.api_client.request(
            "CreateTranscriptionEntity",
            id=transcription.id,
            body=body,
        )

        if self.use_cache:
            # Store transcription entity in local cache
            try:
                CachedTranscriptionEntity.create(
                    transcription=transcription.id,
                    type=tr_entity["type"]["name"],
                    offset=offset,
                    length=length,
                    worker_run_id=self.worker_run_id,
                    confidence=confidence,
                )
            except IntegrityError as e:
                logger.warning(
                    f"Couldn't save created transcription entity in local cache: {e}"
                )

        return tr_entity

    @unsupported_cache
    def create_transcription_entities(
        self,
        transcription: Transcription,
        entities: list[Entity],
    ) -> list[dict[str, str]]:
        """
        Create multiple entities on a transcription in a single API request.

        :param transcription: Transcription to create the entity on.
        :param entities: List of dicts, one per element. Each dict can have the following keys:

            type_id (str)
               Required. ID of the EntityType of the entity.

            length (int)
               Required. Length of the entity in the transcription's text.

            offset (int)
               Required. Starting position of the entity in the transcription's text, as a 0-based index.

            confidence (float or None)
                Optional confidence score, between 0.0 and 1.0.

        :return: List of strings, holding the UUID of each created object.
        """
        assert transcription and isinstance(transcription, Transcription), (
            "transcription shouldn't be null and should be of type Transcription"
        )

        assert entities and isinstance(entities, list), (
            "entities shouldn't be null and should be of type list"
        )

        for index, entity in enumerate(entities):
            assert isinstance(entity, dict), (
                f"Entity at index {index} in entities: Should be of type dict"
            )

            type_id = entity.get("type_id")
            assert type_id and isinstance(type_id, str), (
                f"Entity at index {index} in entities: type_id shouldn't be null and should be of type str"
            )

            offset = entity.get("offset")
            assert offset is not None and isinstance(offset, int) and offset >= 0, (
                f"Entity at index {index} in entities: offset shouldn't be null and should be a positive integer"
            )

            length = entity.get("length")
            assert length is not None and isinstance(length, int) and length > 0, (
                f"Entity at index {index} in entities: length shouldn't be null and should be a strictly positive integer"
            )

            confidence = entity.get("confidence")
            assert confidence is None or (
                isinstance(confidence, float) and 0 <= confidence <= 1
            ), (
                f"Entity at index {index} in entities: confidence should be None or a float in [0..1] range"
            )

        assert len(entities) == len(
            set(map(itemgetter("offset", "length", "type_id"), entities))
        ), "entities should be unique"

        if self.is_read_only:
            logger.warning(
                "Cannot create transcription entities in bulk as this worker is in read-only mode"
            )
            return

        created_tr_entities = self.api_client.request(
            "CreateTranscriptionEntities",
            id=transcription.id,
            body={
                "worker_run_id": self.worker_run_id,
                "transcription_entities": entities,
            },
        )["transcription_entities"]

        return created_tr_entities

    def list_transcription_entities(
        self,
        transcription: Transcription,
        worker_version: str | bool | None = None,
        worker_run: str | bool | None = None,
    ):
        """
        List existing entities on a transcription
        This method does not support cache

        Warns:
        ----
        The following parameters are **deprecated**:

        - `worker_version` in favor of `worker_run`

        :param transcription: The transcription to list entities on.
        :param worker_version: **Deprecated** Restrict to entities created by a worker version with this UUID. Set to False to look for manually created entities.
        :param worker_run: Restrict to entities created by a worker run with this UUID. Set to False to look for manually created entities.
        """
        query_params = {}
        assert transcription and isinstance(transcription, Transcription), (
            "transcription shouldn't be null and should be a Transcription"
        )

        if worker_version is not None:
            warn(
                "`worker_version` usage is deprecated. Consider using `worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(worker_version, str | bool), (
                "worker_version should be of type str or bool"
            )

            if isinstance(worker_version, bool):
                assert worker_version is False, (
                    "if of type bool, worker_version can only be set to False"
                )
            query_params["worker_version"] = worker_version
        if worker_run is not None:
            assert isinstance(worker_run, str | bool), (
                "worker_run should be of type str or bool"
            )
            if isinstance(worker_run, bool):
                assert worker_run is False, (
                    "if of type bool, worker_run can only be set to False"
                )
            query_params["worker_run"] = worker_run

        return self.api_client.paginate(
            "ListTranscriptionEntities", id=transcription.id, **query_params
        )



import logging
import zipfile

import pytest

from arkindex_worker.cache import unsupported_cache
from arkindex_worker.utils import (
    DEFAULT_BATCH_SIZE,
    batch_publication,
    close_delete_file,
    create_zip_archive,
    extract_tar_zst_archive,
    parse_source_id,
)
from tests import FIXTURES_DIR

ARCHIVE = FIXTURES_DIR / "archive.tar.zst"


@pytest.mark.parametrize(
    ("source_id", "expected"),
    [
        (None, None),
        ("", None),
        (
            "cafecafe-cafe-cafe-cafe-cafecafecafe",
            "cafecafe-cafe-cafe-cafe-cafecafecafe",
        ),
        ("manual", False),
    ],
)
def test_parse_source_id(source_id, expected):
    assert parse_source_id(source_id) == expected


def test_extract_tar_zst_archive(tmp_path):
    destination = tmp_path / "destination"
    _, archive_path = extract_tar_zst_archive(ARCHIVE, destination)

    assert archive_path.is_file()
    assert archive_path.suffix == ".tar"
    assert sorted(list(destination.rglob("*"))) == [
        destination / "archive.tar.zst",
        destination / "cache",
        destination / "cache/tables.sqlite",
        destination / "line_transcriptions_small.json",
        destination / "mirrored_image.jpg",
        destination / "page_element.json",
        destination / "rotated_image.jpg",
        destination / "rotated_mirrored_image.jpg",
        destination / "test_image.jpg",
        destination / "tiled_image.jpg",
        destination / "ufcn_line_historical_worker_version.json",
    ]


def test_close_delete_file(tmp_path):
    destination = tmp_path / "destination"
    archive_fd, archive_path = extract_tar_zst_archive(ARCHIVE, destination)
    close_delete_file(archive_fd, archive_path)

    assert not archive_path.exists()


class TestMixin:
    def __init__(self, use_cache: bool = False):
        """
        Args:
            use_cache (bool, optional): To mock BaseWorker.use_cache attribute. Defaults to False.
        """
        self.use_cache = use_cache

    @batch_publication
    def custom_publication_in_batches(self, batch_size: int = DEFAULT_BATCH_SIZE):
        return batch_size

    @unsupported_cache
    @batch_publication
    def custom_publication_in_batches_without_cache(
        self, batch_size: int = DEFAULT_BATCH_SIZE
    ):
        return batch_size


def test_batch_publication_decorator_no_parameter():
    assert TestMixin().custom_publication_in_batches() == DEFAULT_BATCH_SIZE


@pytest.mark.parametrize("wrong_batch_size", [None, "not an int", 0])
def test_batch_publication_decorator_wrong_parameter(wrong_batch_size):
    with pytest.raises(
        AssertionError,
        match="batch_size shouldn't be null and should be a strictly positive integer",
    ):
        TestMixin().custom_publication_in_batches(batch_size=wrong_batch_size)


@pytest.mark.parametrize("batch_size", [1, 10, DEFAULT_BATCH_SIZE])
def test_batch_publication_decorator_right_parameter(batch_size):
    assert (
        TestMixin().custom_publication_in_batches(batch_size=batch_size) == batch_size
    )


def test_batch_publication_decorator_alongside_unsupported_cache(caplog):
    # Capture log messages
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        # Call the helper
        assert (
            TestMixin(use_cache=True).custom_publication_in_batches_without_cache()
            == DEFAULT_BATCH_SIZE
        )

    # Check logs
    assert caplog.record_tuples == [
        (
            "arkindex_worker",
            logging.WARNING,
            "This API helper `custom_publication_in_batches_without_cache` did not update the cache database",
        ),
    ]


def test_zip_archive():
    # Create zip archive from fixtures
    _, archive = create_zip_archive(FIXTURES_DIR / "extract_parent_archives/expected")

    # Check the files in the archive
    with zipfile.ZipFile(archive, mode="r") as f:
        assert sorted(f.namelist()) == [
            "test/",
            "test/images/",
            "test/images/f2649ce7-333e-44d2-ae73-387f18aad1f6.png",
            "test/labels/",
            "test/labels/f2649ce7-333e-44d2-ae73-387f18aad1f6.png",
            "test/labels_json/",
            "test/labels_json/f2649ce7-333e-44d2-ae73-387f18aad1f6.json",
            "train/",
            "train/images/",
            "train/images/98115546-df07-448c-a2f0-34aa24789b77.png",
            "train/images/ebeaa451-9287-4df7-9c40-07eb25cadb78.png",
            "train/labels/",
            "train/labels/98115546-df07-448c-a2f0-34aa24789b77.png",
            "train/labels/ebeaa451-9287-4df7-9c40-07eb25cadb78.png",
            "train/labels_json/",
            "train/labels_json/98115546-df07-448c-a2f0-34aa24789b77.json",
            "train/labels_json/ebeaa451-9287-4df7-9c40-07eb25cadb78.json",
            "val/",
            "val/images/",
            "val/images/2987176d-4338-40f2-90d9-6d2cb4fd4a00.png",
            "val/images/e3f91312-9201-45b7-9c32-e04a97ff1334.png",
            "val/labels/",
            "val/labels/2987176d-4338-40f2-90d9-6d2cb4fd4a00.png",
            "val/labels/e3f91312-9201-45b7-9c32-e04a97ff1334.png",
            "val/labels_json/",
            "val/labels_json/2987176d-4338-40f2-90d9-6d2cb4fd4a00.json",
            "val/labels_json/e3f91312-9201-45b7-9c32-e04a97ff1334.json",
        ]

    # Cleanup
    archive.unlink()



"""Standalone Python worker to create a transcription on Arkindex elements"""

import logging
import os
from argparse import ArgumentParser, Namespace
from typing import Any
from urllib.parse import urljoin

import requests

# Initialize the logger to provide feedback about the worker's execution to the final user
logging.basicConfig(
    format="%(asctime)s %(levelname)s/%(name)s: %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

# Below are listed the environment variables which are mandatory to run this worker
ARKINDEX_API_URL = "ARKINDEX_API_URL"
"""URL that points to the root of the Arkindex instance.
"""
ARKINDEX_API_TOKEN = "ARKINDEX_API_TOKEN"
"""Personal token to authenticate to the Arkindex instance, useful when running locally.
"""
ARKINDEX_TASK_TOKEN = "ARKINDEX_TASK_TOKEN"
"""Machine token to authenticate to the Arkindex instance, useful when running from Arkindex.
"""
ARKINDEX_WORKER_RUN_ID = "ARKINDEX_WORKER_RUN_ID"
"""Identifier to publish worker results.
"""


def parse_args() -> Namespace:
    """Helper to parse command line arguments.
    This worker only supports one optional argument, a list of element IDs to process.

    :return Namespace: A namespace containing the provided command arguments and their value.
    """
    parser = ArgumentParser("python worker.py")
    parser.add_argument(
        "--element",
        nargs="+",
        help="One or more Arkindex element ID",
    )
    return parser.parse_args()


def arkindex_request(
    method: str, endpoint_path: str, body: dict[str, Any] | None = None
) -> dict:
    """Helper to query any endpoint from the Arkindex API.
    The environment variables named `ARKINDEX_API_URL` and `ARKINDEX_API_TOKEN` (or `ARKINDEX_TASK_TOKEN`) are required to use this helper.

    :param str method: The HTTP request method to use https://developer.mozilla.org/en-US/docs/Web/HTTP/Reference/Methods
    :param str endpoint_path: The path of the API endpoint to query
    :param dict[str, Any] | None body: A JSON body to send to the API, defaults to None
    :return dict: The JSON response from the API endpoint
    """
    if body is None:
        body = {}

    # Use the `ARKINDEX_API_URL` environment variable to define the full endpoint URL
    url = urljoin(os.getenv(ARKINDEX_API_URL), endpoint_path)

    # The authorization varies when running locally or in Arkindex
    if "ARKINDEX_TASK_TOKEN" in os.environ:
        authorization = f"Ponos {os.getenv(ARKINDEX_TASK_TOKEN)}"
    else:
        authorization = f"Token {os.getenv(ARKINDEX_API_TOKEN)}"

    # Query the endpoint URL using the `requests` Python package
    response = requests.request(
        method=method,
        url=url,
        headers={"Authorization": authorization},
        json=body,
    )

    # Raise an exception if anything went wrong while querying the endpoint
    try:
        response.raise_for_status()
    except requests.HTTPError:
        logger.error(
            f"Request `{endpoint_path}` failed with code {response.status_code}: {response.content}"
        )
        raise

    # Return the response in JSON format if it was successful
    return response.json()


def main() -> None:
    """Standalone Python worker to create a transcription on Arkindex elements"""
    # Check that the required environment variables are available
    for variable in (ARKINDEX_API_URL, ARKINDEX_WORKER_RUN_ID):
        assert os.getenv(variable), (
            f"Missing required variable `{variable}` in the environment."
        )

    assert os.getenv(ARKINDEX_API_TOKEN) or os.getenv(ARKINDEX_TASK_TOKEN), (
        f"Either `{ARKINDEX_API_TOKEN}` or `{ARKINDEX_TASK_TOKEN}` variable must be set in the environment."
    )

    # Retrieve the worker configuration from Arkindex
    # API endpoint: https://arkindex.teklia.com/api-docs/#tag/process/operation/RetrieveWorkerRun
    configuration = arkindex_request(
        method="get",
        endpoint_path=f"process/workers/{os.getenv(ARKINDEX_WORKER_RUN_ID)}/",
    )

    # Build the list of elements to process
    elements = []

    # Option 1: The worker is running locally, on your machine, we use the value of the `--element` command argument
    if configuration["process"]["mode"] == "local":
        # Parse the provided command arguments
        args = parse_args()

        # Retrieve the list of elements from the `--element` argument
        elements = args.element

        # Assert that at least one element was provided to run the worker on
        assert elements, (
            "Missing at least one element ID to process while running the worker locally."
        )

    # Option 2: The worker is running on Arkindex, in a process, we list process elements
    else:
        # Retrieve the list of elements from the process which is currently running
        # API endpoint: https://arkindex.teklia.com/api-docs/#tag/process/operation/ListProcessElements
        json_response = arkindex_request(
            method="get",
            endpoint_path=f"process/{configuration['process']['id']}/elements/",
        )

        # We only need the ID of each element to process, other information is not necessary
        elements = [element["id"] for element in json_response["results"]]

    total = len(elements)
    failed = 0
    # Iterate over all elements to create a basic transcription
    for element_id in elements:
        try:
            # Create the "Hello world!" transcription on the current element
            # API endpoint: https://arkindex.teklia.com/api-docs/#tag/transcriptions/operation/CreateTranscription
            transcription = arkindex_request(
                method="post",
                endpoint_path=f"element/{element_id}/transcription/",
                body={
                    "text": "Hello world!",
                    "worker_run_id": os.getenv(ARKINDEX_WORKER_RUN_ID),
                    "confidence": 1.0,
                },
            )

            # Output feedback when a transcription is successfully created
            logger.info(
                f"A transcription with the ID {transcription['id']} was successfully created on element {element_id}."
            )

        except Exception:
            # Output feedback when failing to create a transcription, and increment the `failed` counter
            logger.error(f"Failed to create a transcription on element {element_id}.")
            failed += 1

    completed = total - failed
    # Output a summary of the worker execution over all provided elements
    logger.info(f"Ran on {total} element(s): {completed} completed, {failed} error(s).")


if __name__ == "__main__":
    main()
